# Copyright (c) 2014, Chris Church <chris@ninemoreminutes.com>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

DOCUMENTATION = """
name: powershell
version_added: historical
short_description: Windows PowerShell
description:
- The only option when using 'winrm' or 'psrp' as a connection plugin.
- Can also be used when using 'ssh' as a connection plugin and the C(DefaultShell) has been configured to PowerShell.
extends_documentation_fragment:
- shell_windows
"""

import base64
import os
import shlex
import ntpath

from ansible._internal._powershell import _script
from ansible.executor.powershell.module_manifest import _bootstrap_powershell_script
from ansible.module_utils.common.text.converters import to_text
from ansible.plugins.shell import ShellBase, _ShellCommand
from ansible.utils.display import Display


display = Display()

_common_args = ['PowerShell', '-NoProfile', '-NonInteractive', '-ExecutionPolicy', 'Unrestricted']


class ShellModule(ShellBase):

    # Common shell filenames that this plugin handles
    # Powershell is handled differently.  It's selected when winrm is the
    # connection
    COMPATIBLE_SHELLS = frozenset()  # type: frozenset[str]
    # Family of shells this has.  Must match the filename without extension
    SHELL_FAMILY = 'powershell'

    # We try catch as some connection plugins don't have a console (PSRP).
    _CONSOLE_ENCODING = "try { [Console]::OutputEncoding = New-Object System.Text.UTF8Encoding } catch {}"
    _SHELL_REDIRECT_ALLNULL = '> $null'
    _SHELL_AND = ';'

    # Used by various parts of Ansible to do Windows specific changes
    _IS_WINDOWS = True

    # TODO: add binary module support

    def env_prefix(self, **kwargs):
        # powershell/winrm env handling is handled in the exec wrapper
        return ""

    def join_path(self, *args):
        # use normpath() to remove doubled slashed and convert forward to backslashes
        parts = [ntpath.normpath(arg) for arg in args]

        # Because ntpath.join treats any component that begins with a backslash as an absolute path,
        # we have to strip slashes from at least the beginning, otherwise join will ignore all previous
        # path components except for the drive.
        return ntpath.join(parts[0], *[part.strip('\\') for part in parts[1:]])

    def get_remote_filename(self, pathname):
        # powershell requires that script files end with .ps1
        base_name = os.path.basename(pathname.strip())
        name, ext = os.path.splitext(base_name.strip())
        if ext.lower() not in ['.ps1', '.exe']:
            return name + '.ps1'

        return base_name.strip()

    def path_has_trailing_slash(self, path):
        # Allow Windows paths to be specified using either slash.
        return path.endswith('/') or path.endswith('\\')

    def chmod(self, paths, mode):
        raise NotImplementedError('chmod is not implemented for Powershell')

    def chown(self, paths, user):
        raise NotImplementedError('chown is not implemented for Powershell')

    def set_user_facl(self, paths, user, mode):
        raise NotImplementedError('set_user_facl is not implemented for Powershell')

    def remove(self, path, recurse=False):
        path = _script.quote_pwsh_argument(path)
        script = f"Remove-Item {path} -Force"
        if recurse:
            script += " -Recurse"

        return self._encode_pwsh_script_as_command(script)

    def mkdtemp(
        self,
        basefile: str | None = None,
        system: bool = False,
        mode: int = 0o700,
        tmpdir: str | None = None,
    ) -> str:
        # This is used when connection plugins do not support pipelining.
        if not basefile:
            basefile = self.__class__._generate_temp_dir_name()
        basefile = _script.quote_pwsh_argument(basefile, force_quote=True)
        basetmpdir = _script.quote_pwsh_argument(tmpdir if tmpdir else self.get_option('remote_tmp'), force_quote=True)

        script = f"""
        {self._CONSOLE_ENCODING}
        $tmp_path = [System.Environment]::ExpandEnvironmentVariables({basetmpdir})
        $tmp = New-Item -Type Directory -Path $tmp_path -Name {basefile}
        Write-Output -InputObject $tmp.FullName
        """
        return self._encode_pwsh_script_as_command(script)

    def _mkdtemp2(
        self,
        basefile: str | None = None,
        system: bool = False,
        mode: int = 0o700,
        tmpdir: str | None = None,
    ) -> _ShellCommand:
        # Windows does not have an equivalent for the system temp files, so
        # the param is ignored
        if not basefile:
            basefile = self.__class__._generate_temp_dir_name()

        basetmpdir = tmpdir if tmpdir else self.get_option('remote_tmp')

        script, stdin = _bootstrap_powershell_script("powershell_mkdtemp.ps1", {
            'Directory': basetmpdir,
            'Name': basefile,
        })
        command = self._encode_pwsh_script_as_command(script)

        return _ShellCommand(
            command=command,
            input_data=stdin,
        )

    def expand_user(
        self,
        user_home_path: str,
        username: str = '',
    ) -> str:
        # This is used when connection plugins do not support pipelining.

        script = self._CONSOLE_ENCODING
        if user_home_path == '~':
            script += "; (Get-Location).Path"
        elif user_home_path.startswith('~\\'):
            quoted_home = _script.quote_pwsh_argument(user_home_path[1:], force_quote=True)
            script += f"; ((Get-Location).Path + {quoted_home})"
        else:
            script += f"; {_script.quote_pwsh_argument(user_home_path, force_quote=True)}"

        return self._encode_pwsh_script_as_command(script)

    def _expand_user2(
        self,
        user_home_path: str,
        username: str = '',
    ) -> _ShellCommand:
        script, stdin = _bootstrap_powershell_script("powershell_expand_user.ps1", {
            'Path': user_home_path,
        })
        command = self._encode_pwsh_script_as_command(script)

        return _ShellCommand(
            command=command,
            input_data=stdin,
        )

    def exists(self, path):
        path = _script.quote_pwsh_argument(path)
        script = f"exit -not (Test-Path -LiteralPath {path})"

        return self._encode_pwsh_script_as_command(script)

    def checksum(self, path, *args, **kwargs):
        display.deprecated(
            msg="The `ShellModule.checksum` method is deprecated.",
            version="2.23",
            help_text="Use `ActionBase._execute_remote_stat()` instead.",
        )
        path = _script.quote_pwsh_argument(path)
        script = """
            If (Test-Path -PathType Leaf %(path)s)
            {
                $sp = new-object -TypeName System.Security.Cryptography.SHA1CryptoServiceProvider;
                $fp = [System.IO.File]::Open(%(path)s, [System.IO.Filemode]::Open, [System.IO.FileAccess]::Read);
                [System.BitConverter]::ToString($sp.ComputeHash($fp)).Replace("-", "").ToLower();
                $fp.Dispose();
            }
            ElseIf (Test-Path -PathType Container %(path)s)
            {
                Write-Output "3";
            }
            Else
            {
                Write-Output "1";
            }
        """ % dict(path=path)

        return self._encode_pwsh_script_as_command(script)

    def build_module_command(self, env_string, shebang, cmd, arg_path=None):
        # This will be called when executing modules not covered by the pwsh
        # exec wrapper. For example module replacer, binary or basic shell
        # modules.
        # The env_string is never used in PowerShell.
        cmd_parts = shlex.split(cmd, posix=False)

        if shebang and shebang.startswith('#!'):
            cmd = shebang[2:]
        elif not shebang:
            # The module is assumed to be a binary
            cmd = cmd_parts.pop(0)

        if arg_path:
            cmd_parts.append(arg_path)

        script = _script.build_pwsh_cmd_statement(cmd, cmd_parts)
        return self._encode_pwsh_script_as_command(script)

    def wrap_for_exec(self, cmd):
        super().wrap_for_exec(cmd)
        return '& %s; exit $LASTEXITCODE' % cmd

    def join(self, cmd_parts: list[str]) -> str:
        if not cmd_parts:
            return ""

        cmd = cmd_parts[0]
        args = []
        if len(cmd_parts) > 1:
            args = cmd_parts[1:]

        return _script.build_pwsh_cmd_statement(cmd, args)

    def quote(self, cmd: str) -> str:
        return _script.quote_pwsh_argument(cmd)

    def _encode_pwsh_script_as_command(self, script: str) -> str:
        """Wraps the provided PowerShell script as a command line string for the current shell plugin."""
        cmd_args = _script.get_pwsh_encoded_cmdline(script, override_execution_policy=True)
        return self.join(cmd_args)

    def _encode_script(self, script, as_list=False, strict_mode=True, preserve_rc=True):
        """Convert a PowerShell script to a single base64-encoded command."""
        # Needs a deprecation as ansible.windows uses this in the reboot plugin.
        display.deprecated(
            msg="The `PowerShell._encode_script` method is deprecated.",
            version="2.24",
            help_text="Contact plugin author to update their plugin to use an alternative method.",
        )
        script = to_text(script)

        if script == u'-':
            cmd_parts = _common_args + ['-Command', '-']

        else:
            if strict_mode:
                script = u'Set-StrictMode -Version Latest\r\n%s' % script
            # try to propagate exit code if present- won't work with begin/process/end-style scripts (ala put_file)
            # NB: the exit code returned may be incorrect in the case of a successful command followed by an invalid command
            if preserve_rc:
                script = u'%s\r\nIf (-not $?) { If (Get-Variable LASTEXITCODE -ErrorAction SilentlyContinue) { exit $LASTEXITCODE } Else { exit 1 } }\r\n'\
                    % script
            script = '\n'.join([x.strip() for x in script.splitlines() if x.strip()])
            encoded_script = to_text(base64.b64encode(script.encode('utf-16-le')), 'utf-8')
            cmd_parts = _common_args + ['-EncodedCommand', encoded_script]

        if as_list:
            return cmd_parts
        return ' '.join(cmd_parts)



# -*- coding: utf-8 -*-

# (c) 2012, Michael DeHaan <michael.dehaan@gmail.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = r"""
---
module: slurp
version_added: historical
short_description: Slurps a file from remote nodes
description:
     - This module works like M(ansible.builtin.fetch). It is used for fetching a base64-
       encoded blob containing the data in a remote file.
     - This module is also supported for Windows targets.
options:
  src:
    description:
      - The file on the remote system to fetch. This I(must) be a file, not a directory.
    type: path
    required: true
    aliases: [ path ]
  armor:
    description:
     - To safely deliver the data requested, armor it by base64 encoding it.
     - This is mainly done to avoid issues with binary data or to preserve non-UTF-8 encodings,
        as the content will be converted to UTF-8 due to the JSON transport used.
     - For windows support, check the ansible.windows collection.
    type: bool
    default: true
    version_added: '2.21'
extends_documentation_fragment:
    - action_common_attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  platform:
    platforms: posix, windows
notes:
   - This module returns an 'in memory' base64 encoded version of the file, take
     into account that this will require at least twice the RAM as the original file size.
seealso:
- module: ansible.builtin.fetch
author:
    - Ansible Core Team
    - Michael DeHaan (@mpdehaan)
"""

EXAMPLES = r"""
- name: Find out what the remote machine's mounts are
  ansible.builtin.slurp:
    src: /proc/mounts
  register: mounts

- name: Print returned information
  ansible.builtin.debug:
    msg: "{{ mounts['content'] | b64decode }}"

- name: Find out what the remote machine's mounts are, no armor
  ansible.builtin.slurp:
    src: /proc/mounts
    armor: false
  register: mounts

- name: Print returned information
  ansible.builtin.debug:
    msg: "{{ mounts['content'] }}"

# From the commandline, find the pid of the remote machine's sshd
# $ ansible host -m ansible.builtin.slurp -a 'src=/var/run/sshd.pid'
# host | SUCCESS => {
#     "changed": false,
#     "content": "MjE3OQo=",
#     "encoding": "base64",
#     "source": "/var/run/sshd.pid"
# }
# $ echo MjE3OQo= | base64 -d
# 2179
"""

RETURN = r"""
content:
    description: Encoded file content
    returned: success
    type: str
    sample: "MjE3OQo="
encoding:
    description: Current encoding of the file content, it can be C(base64)
                 or C(UTF-8) depending on the value of the O(armor) option.
    returned: success
    type: str
    sample: "base64"
source:
    description: Actual path of file slurped
    returned: success
    type: str
    sample: "/var/run/sshd.pid"
"""

import base64
import errno

from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(
            src=dict(type='path', required=True, aliases=['path']),
            armor=dict(type='bool', default=True),
        ),
        supports_check_mode=True,
    )
    source = module.params['src']
    armor = module.params['armor']

    try:
        with open(source, 'rb') as source_fh:
            if armor:
                encoding = 'base64'
                data = base64.b64encode(source_fh.read())
            else:
                # not current file encoding, but will be once passed as JSON
                encoding = 'utf-8'
                data = source_fh.read()
    except OSError as ex:
        if ex.errno == errno.ENOENT:
            msg = f"File not found: {source}"
        elif ex.errno == errno.EACCES:
            msg = f"File is not readable: {source}"
        elif ex.errno == errno.EISDIR:
            msg = f"Source is a directory and must be a file: {source}"
        else:
            msg = "Unable to slurp file: {source}"

        module.fail_json(msg, exception=ex)

    module.exit_json(content=data, source=source, encoding=encoding)


if __name__ == '__main__':
    main()



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import ctypes
import ctypes.util
import os
import sys
import typing as t
from dataclasses import dataclass

__all__ = ['CryptFacade']

_FAILURE_TOKENS = frozenset({b'*', b'*0', b'*1'})


@dataclass(frozen=True)
class _CryptLib:
    name: str | None
    exclude_platforms: frozenset[str] = frozenset()
    include_platforms: frozenset[str] = frozenset()
    is_path: bool = False


_CRYPT_LIBS = (
    _CryptLib('crypt'),  # libxcrypt
    _CryptLib(None, exclude_platforms=frozenset({'darwin'})),  # fallback to default libc
    _CryptLib(  # macOS Homebrew (Apple Silicon)
        '/opt/homebrew/opt/libxcrypt/lib/libcrypt.dylib',
        include_platforms=frozenset({'darwin'}),
        is_path=True,
    ),
    _CryptLib(  # macOS Homebrew (Intel)
        '/usr/local/opt/libxcrypt/lib/libcrypt.dylib',
        include_platforms=frozenset({'darwin'}),
        is_path=True,
    ),
)


class CryptFacade:
    """
    Provide an interface for various crypt libraries that might be available.
    """

    def __init__(self) -> None:
        self._crypt_impl: t.Callable | None = None
        self._crypt_gensalt_impl: t.Callable | None = None
        self._use_crypt_r = False
        self._use_crypt_gensalt_rn = False
        self._crypt_name = ""

        self._setup()

    class _CryptData(ctypes.Structure):
        _fields_ = [('_opaque', ctypes.c_char * 131072)]

    @property
    def has_crypt_gensalt(self) -> bool:
        return self._crypt_gensalt_impl is not None

    def _setup(self) -> None:
        """Setup crypt implementation"""
        for lib_config in _CRYPT_LIBS:
            if sys.platform in lib_config.exclude_platforms:
                continue
            if lib_config.include_platforms and sys.platform not in lib_config.include_platforms:
                continue

            if lib_config.name is None:
                lib_so = None
            elif lib_config.is_path:
                if os.path.exists(lib_config.name):
                    lib_so = lib_config.name
                else:
                    continue
            else:
                lib_so = ctypes.util.find_library(lib_config.name)
                if not lib_so:
                    continue

            loaded_lib = ctypes.cdll.LoadLibrary(lib_so)

            try:
                self._crypt_impl = loaded_lib.crypt_r
                self._use_crypt_r = True
            except AttributeError:
                try:
                    self._crypt_impl = loaded_lib.crypt
                except AttributeError:
                    continue

            if self._use_crypt_r:
                self._crypt_impl.argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(self._CryptData)]
            else:
                self._crypt_impl.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
            self._crypt_impl.restype = ctypes.c_char_p

            # Try to load crypt_gensalt (available in libxcrypt)
            try:
                self._crypt_gensalt_impl = loaded_lib.crypt_gensalt_rn
                self._crypt_gensalt_impl.argtypes = [ctypes.c_char_p, ctypes.c_ulong, ctypes.c_char_p, ctypes.c_int, ctypes.c_char_p, ctypes.c_int]
                self._crypt_gensalt_impl.restype = ctypes.c_char_p
                self._use_crypt_gensalt_rn = True
            except AttributeError:
                try:
                    self._crypt_gensalt_impl = loaded_lib.crypt_gensalt
                    self._crypt_gensalt_impl.argtypes = [ctypes.c_char_p, ctypes.c_ulong, ctypes.c_char_p, ctypes.c_int]
                    self._crypt_gensalt_impl.restype = ctypes.c_char_p
                except AttributeError:
                    self._crypt_gensalt_impl = None

            self._crypt_name = lib_config.name
            break
        else:
            raise ImportError('Cannot find crypt implementation')

    def crypt(self, word: bytes, salt: bytes) -> bytes:
        """Hash a password using the system's crypt function."""
        ctypes.set_errno(0)

        if self._use_crypt_r:
            data = self._CryptData()
            ctypes.memset(ctypes.byref(data), 0, ctypes.sizeof(data))
            result = self._crypt_impl(word, salt, ctypes.byref(data))
        else:
            result = self._crypt_impl(word, salt)

        errno = ctypes.get_errno()
        if errno:
            error_msg = os.strerror(errno)
            raise OSError(errno, f'crypt failed: {error_msg}')

        if result is None or result in _FAILURE_TOKENS:
            raise ValueError('crypt failed: invalid salt or unsupported algorithm')

        return result

    def crypt_gensalt(self, prefix: bytes, count: int, rbytes: bytes) -> bytes:
        """Generate a salt string for use with crypt."""
        if not self.has_crypt_gensalt:
            raise NotImplementedError('crypt_gensalt not available (requires libxcrypt)')

        ctypes.set_errno(0)

        if self._use_crypt_gensalt_rn:
            output = ctypes.create_string_buffer(256)
            result = self._crypt_gensalt_impl(prefix, count, rbytes, len(rbytes), output, len(output))
            if result is not None:
                result = output.value
        else:
            result = self._crypt_gensalt_impl(prefix, count, rbytes, len(rbytes))

        errno = ctypes.get_errno()
        if errno:
            error_msg = os.strerror(errno)
            raise OSError(errno, f'crypt_gensalt failed: {error_msg}')

        if result is None:
            raise ValueError('crypt_gensalt failed: unable to generate salt')

        if result in _FAILURE_TOKENS:
            raise ValueError('crypt_gensalt failed: invalid prefix or unsupported algorithm')

        return result



"""A plugin for pylint to identify imports and functions which should not be used."""

from __future__ import annotations

import functools
import os
import typing as t

import astroid.exceptions
import astroid.nodes

from pylint.checkers import BaseChecker

ANSIBLE_TEST_MODULES_PATH = os.environ['ANSIBLE_TEST_MODULES_PATH']
ANSIBLE_TEST_MODULE_UTILS_PATH = os.environ['ANSIBLE_TEST_MODULE_UTILS_PATH']


class UnwantedEntry:
    """Defines an unwanted import."""
    def __init__(
            self,
            alternative,  # type: str
            modules_only=False,  # type: bool
            names=None,  # type: t.Optional[t.Tuple[str, ...]]
            ignore_paths=None,  # type: t.Optional[t.Tuple[str, ...]]
            ansible_test_only=False,  # type: bool
    ):  # type: (...) -> None
        self.alternative = alternative
        self.modules_only = modules_only
        self.names = set(names) if names else set()
        self.ignore_paths = ignore_paths
        self.ansible_test_only = ansible_test_only

    def applies_to(self, path, name=None):  # type: (str, t.Optional[str]) -> bool
        """Return True if this entry applies to the given path, otherwise return False."""
        if self.names:
            if not name:
                return False

            if name not in self.names:
                return False

        if self.ignore_paths and any(path.endswith(ignore_path) for ignore_path in self.ignore_paths):
            return False

        if self.ansible_test_only and '/test/lib/ansible_test/_internal/' not in path:
            return False

        if self.modules_only:
            return is_module_path(path)

        return True


def is_module_path(path):  # type: (str) -> bool
    """Return True if the given path is a module or module_utils path, otherwise return False."""
    return path.startswith(ANSIBLE_TEST_MODULES_PATH) or path.startswith(ANSIBLE_TEST_MODULE_UTILS_PATH)


class AnsibleUnwantedChecker(BaseChecker):
    """Checker for unwanted imports and functions."""

    name = 'unwanted'

    BAD_IMPORT = 'ansible-bad-import'
    BAD_IMPORT_FROM = 'ansible-bad-import-from'
    BAD_FUNCTION = 'ansible-bad-function'
    BAD_MODULE_IMPORT = 'ansible-bad-module-import'

    msgs = dict(
        E5101=('Import %s instead of %s',
               BAD_IMPORT,
               'Identifies imports which should not be used.'),
        E5102=('Import %s from %s instead of %s',
               BAD_IMPORT_FROM,
               'Identifies imports which should not be used.'),
        E5103=('Call %s instead of %s',
               BAD_FUNCTION,
               'Identifies functions which should not be used.'),
        E5104=('Import external package or ansible.module_utils not %s',
               BAD_MODULE_IMPORT,
               'Identifies imports which should not be used.'),
    )

    unwanted_imports = {
        # see https://docs.python.org/3/library/collections.abc.html
        # deprecated: description='remove collections check now that Python 3.9 is no longer supported' core_version='2.23'
        'collections': UnwantedEntry(
            'collections.abc',
            names=(
                'MappingView',
                'ItemsView',
                'KeysView',
                'ValuesView',
                'Mapping', 'MutableMapping',
                'Sequence', 'MutableSequence',
                'Set', 'MutableSet',
                'Container',
                'Hashable',
                'Sized',
                'Callable',
                'Iterable',
                'Iterator',
            )
        ),

        'ansible.module_utils.six': UnwantedEntry(
            'the Python standard library equivalent'
        ),
    }

    unwanted_functions = {
        # see https://docs.python.org/3/library/tempfile.html#tempfile.mktemp
        'tempfile.mktemp': UnwantedEntry('tempfile.mkstemp'),

        'os.chmod': UnwantedEntry('verified_chmod', ansible_test_only=True),

        'sys.exit': UnwantedEntry('exit_json or fail_json',
                                  ignore_paths=(
                                      '/lib/ansible/module_utils/basic.py',
                                      '/lib/ansible/modules/async_wrapper.py',
                                  ),
                                  modules_only=True),

        'builtins.print': UnwantedEntry('module.log or module.debug',
                                        ignore_paths=(
                                            '/lib/ansible/module_utils/basic.py',
                                        ),
                                        modules_only=True),
    }

    for method in (
        'popen',
        'posix_spawn',
        'posix_spawnp',
        'spawnl',
        'spawnle',
        'spawnlp',
        'spawnlpe',
        'spawnv',
        'spawnve',
        'spawnvp',
        'spawnvpe',
        'system',
    ):
        unwanted_functions[f'os.{method}'] = UnwantedEntry('run_command', modules_only=True)

    for method in (
        'Popen',
        'call',
        'check_call',
        'check_output',
        'getoutput',
        'getstatusoutput',
        'run',
    ):
        unwanted_functions[f'subprocess.{method}'] = UnwantedEntry('run_command', modules_only=True)

    @functools.cached_property
    def is_ansible_core(self) -> bool:
        """True if ansible-core is being tested."""
        return not self.linter.config.collection_name

    def visit_import(self, node: astroid.nodes.Import) -> None:
        """Visit an import node."""
        for name in node.names:
            self._check_import(node, name[0])

    def visit_importfrom(self, node: astroid.nodes.ImportFrom) -> None:
        """Visit an import from node."""
        self._check_importfrom(node, node.modname, node.names)

    def visit_attribute(self, node: astroid.nodes.Attribute) -> None:
        """Visit an attribute node."""
        last_child = node.last_child()

        # this is faster than using type inference and will catch the most common cases
        if not isinstance(last_child, astroid.nodes.Name):
            return

        module = last_child.name

        entry = self.unwanted_imports.get(module)

        if entry and entry.names:
            if entry.applies_to(self.linter.current_file, node.attrname):
                self.add_message(self.BAD_IMPORT_FROM, args=(node.attrname, entry.alternative, module), node=node)

    def visit_call(self, node: astroid.nodes.Call) -> None:
        """Visit a call node."""
        try:
            for i in node.func.inferred():
                func = None

                if isinstance(i.parent, astroid.nodes.Module):
                    parent_module = i.parent.name
                elif isinstance(i.parent, astroid.nodes.If) and isinstance(i.parent.parent, astroid.nodes.Module):
                    parent_module = i.parent.parent.name
                else:
                    parent_module = None

                if parent_module == 'posix':
                    parent_module = 'os'  # some os.* functions we're looking for show up as posix.* imports

                if parent_module and isinstance(i, (astroid.nodes.FunctionDef, astroid.nodes.ClassDef)):
                    func = f'{parent_module}.{i.name}'

                if not func:
                    continue

                entry = self.unwanted_functions.get(func)

                if entry and entry.applies_to(self.linter.current_file):
                    self.add_message(self.BAD_FUNCTION, args=(entry.alternative, func), node=node)
        except astroid.exceptions.InferenceError:
            pass

    def _check_import(self, node: astroid.nodes.Import, modname: str) -> None:
        """Check the imports on the specified import node."""
        self._check_module_import(node, modname)

        entry = self.unwanted_imports.get(modname)

        if not entry:
            return

        if entry.applies_to(self.linter.current_file):
            self.add_message(self.BAD_IMPORT, args=(entry.alternative, modname), node=node)

    def _check_importfrom(self, node: astroid.nodes.ImportFrom, modname: str, names: list[tuple[str, str | None]]) -> None:
        """Check the imports on the specified import from node."""
        self._check_module_import(node, modname)

        entry = self.unwanted_imports.get(modname)

        if not entry:
            return

        for name in names:
            if entry.applies_to(self.linter.current_file, name[0]):
                self.add_message(self.BAD_IMPORT_FROM, args=(name[0], entry.alternative, modname), node=node)

    def _check_module_import(self, node: astroid.nodes.Import | astroid.nodes.ImportFrom, modname: str) -> None:
        """Check the module import on the given import or import from node."""
        if not is_module_path(self.linter.current_file):
            return

        if modname == 'ansible.module_utils' or modname.startswith('ansible.module_utils.'):
            return

        if modname == 'ansible' or modname.startswith('ansible.'):
            self.add_message(self.BAD_MODULE_IMPORT, args=(modname,), node=node)


def register(linter):
    """required method to auto register this checker """
    linter.register_checker(AnsibleUnwantedChecker(linter))



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import itertools
import os
import time
import json
import pathlib
import subprocess
import sys

import traceback
import typing as t

from ansible import constants as C
from ansible._internal._task import TaskContext, UnifiedTaskResult
from ansible.cli import scripts
from ansible.errors import (
    AnsibleError, AnsibleParserError, AnsibleUndefinedVariable, AnsibleTaskError,
    AnsibleValueOmittedError,
)

from ansible._internal import _display_utils
from ansible._internal._datatag import _utils
from ansible.module_utils.datatag import native_type_name
from ansible._internal._datatag._tags import TrustedAsTemplate
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.module_utils.common.text.converters import to_text, to_native
from ansible.module_utils.connection import write_to_stream
from ansible.playbook.play_context import PlayContext
from ansible.plugins import get_plugin_class
from ansible.plugins.action import ActionBase
from ansible.plugins.connection import ConnectionBase
from ansible.plugins.loader import become_loader, cliconf_loader, connection_loader, httpapi_loader, netconf_loader, terminal_loader, PluginLoadContext
from ansible._internal._templating._jinja_plugins import _invoke_lookup, _DirectCall
from ansible._internal._templating._engine import TemplateEngine
from ansible.template import Templar
from ansible.utils.collection_loader import AnsibleCollectionConfig
from ansible.utils.display import Display
from ansible.utils.vars import combine_vars
from ansible.vars.clean import namespace_facts
from ansible.vars.manager import _clean_and_deprecate_top_level_facts, _INJECT_FACTS
from ansible._internal._errors import _task_timeout
from ansible._internal import _task
from ansible.playbook.task import Task
from ansible.executor.task_queue_manager import FinalQueue


display = Display()


_DELEGATED_CONNECTION_PLUGIN_VAR_NAMES = frozenset({
    'ansible_host',
    'ansible_port',
    'ansible_user',
    'ansible_connection',
})

__all__ = ['TaskExecutor']


class TaskExecutor:

    """
    This is the main worker class for the executor pipeline, which
    handles loading an action plugin to actually dispatch the task to
    a given host. This class roughly corresponds to the old Runner()
    class.
    """

    def __init__(
        self,
        host,
        play_context: PlayContext,
        loader,
        shared_loader_obj,
        final_q: FinalQueue,
        variable_manager,
    ) -> None:
        self._host = host
        self._play_context = play_context
        self._loader = loader
        self._shared_loader_obj = shared_loader_obj
        self._connection = None
        self._final_q = final_q
        self._variable_manager = variable_manager
        self._loop_eval_error: Exception | None = None

    @property
    def _task(self) -> Task:
        return TaskContext.current().task

    def run(self) -> UnifiedTaskResult:
        """
        The main executor entrypoint, where we determine if the specified
        task requires looping and either runs the task with self._run_loop()
        or self._execute(). After that, the returned results are parsed and
        returned as a dict.
        """

        display.debug("in run() - task %s" % self._task._uuid)

        task_ctx = TaskContext.current()

        try:
            try:
                items = self._get_loop_items()
            except AnsibleUndefinedVariable as e:
                # save the error raised here for use later
                items = None
                self._loop_eval_error = e

            if items is None:  # non-loop
                utr = self._execute()
            elif not items:  # empty loop
                with UnifiedTaskResult.create_and_record() as utr:
                    utr.set_skipped('No items in the list', include_skipped_reason=True)
                    utr.loop_results = []
            else:  # loop with values
                utr = self._run_loop(items)

            return utr
        except Exception as ex:
            utr = UnifiedTaskResult.create_from_action_exception(ex)

            self._task.update_result_no_log(task_ctx.task_templar, utr)

            if not isinstance(ex, AnsibleError):
                utr.msg = f'Unexpected failure during task execution: {utr.msg}'

            return utr
        finally:
            try:
                self._connection.close()
            except AttributeError:
                pass
            except Exception as e:
                display.debug(u"error closing connection: %s" % to_text(e))

    def _get_loop_items(self) -> list[t.Any] | None:
        """
        Loads a lookup plugin to handle the with_* portion of a task (if specified),
        and returns the items result.
        """
        task_ctx = TaskContext.current()

        # get search path for this task to pass to lookup plugins
        task_ctx.task_vars['ansible_search_path'] = self._task.get_search_path()

        # ensure basedir is always in (dwim already searches here but we need to display it)
        if self._loader.get_basedir() not in task_ctx.task_vars['ansible_search_path']:
            task_ctx.task_vars['ansible_search_path'].append(self._loader.get_basedir())

        items = None
        if self._task.loop_with:
            terms = self._task.loop

            if isinstance(terms, str):
                terms = task_ctx.task_templar.resolve_to_container(_utils.str_problematic_strip(terms))

            if not isinstance(terms, list):
                terms = [terms]

            @_DirectCall.mark
            def invoke_lookup() -> t.Any:
                """Scope-capturing wrapper around _invoke_lookup to avoid functools.partial obscuring its usage from type-checking tools."""
                return _invoke_lookup(
                    plugin_name=self._task.loop_with,
                    lookup_terms=terms,
                    lookup_kwargs=dict(wantlist=True),
                    invoked_as_with=True,
                )

            # Smuggle a special wrapped lookup invocation in as a local variable for its exclusive use when being evaluated as `with_(lookup)`.
            # This value will not be visible to other users of this templar or its `available_variables`.
            items = task_ctx.task_templar.evaluate_expression(
                expression=TrustedAsTemplate().tag("invoke_lookup()"),
                local_variables=dict(invoke_lookup=invoke_lookup),
            )

        elif self._task.loop is not None:
            items = task_ctx.task_templar.template(self._task.loop)

            if not isinstance(items, list):
                raise AnsibleError(
                    f"The `loop` value must resolve to a 'list', not {native_type_name(items)!r}.",
                    help_text="Provide a list of items/templates, or a template resolving to a list.",
                    obj=self._task.loop,
                )

        return items

    def _run_loop(self, items: list[t.Any]) -> UnifiedTaskResult:
        """
        Runs the task with the loop items specified and collates the result
        into an array named 'results' which is inserted into the final result
        along with the item for which the loop ran.
        """
        task_ctx = TaskContext.current()

        task_ctx._loop_items = items

        self._task.loop_control.post_validate(templar=task_ctx.task_templar)
        self._check_loop_control()

        loop_pause = self._task.loop_control.pause

        ran_once = False
        last_loop_task: Task | None = None

        for _item_index, _item in task_ctx.start_loop():
            # pause between loop iterations
            if loop_pause and ran_once:
                time.sleep(loop_pause)
            else:
                ran_once = True

            with task_ctx.loop_item_context(self):
                last_loop_task = self._task

                utr = self._execute()

                if self._task.loop_control and self._task.loop_control.break_when:
                    try:
                        utr.set_break_when_result(self._task._resolve_conditional(self._task.loop_control.break_when, task_ctx.task_vars))
                    except AnsibleError as ex:
                        # RPFIX-5: UX: This this bypasses AnsibleTaskError handling, resulting in less information than a normal task failure.
                        utr.set_break_when_result(ex)

                self._update_task_connection()

                if utr.failed or utr.unreachable:
                    self._final_q.send_callback('v2_runner_item_on_failed', self._host, self._task, utr)
                elif utr.skipped:
                    self._final_q.send_callback('v2_runner_item_on_skipped', self._host, self._task, utr)
                else:
                    if self._task.diff:
                        # non-loop diff dispatch is handled controller-side by the strategy
                        self._final_q.send_callback('v2_on_file_diff', self._host, self._task, utr)

                    self._final_q.send_callback('v2_runner_item_on_ok', self._host, self._task, utr)

            # update the connection value on the original task to reflect the resolved value
            self._update_task_connection()

        if last_loop_task:
            # FUTURE: hide this in Task/LoopContext once they're fully implemented
            # NOTE: run_once cannot contain loop vars because it's templated earlier also
            # This is saving the post-validated field from the last loop so the strategy can use the templated value post task execution
            self._task.run_once = last_loop_task.run_once
            self._task.action = last_loop_task.action

        return task_ctx.build_loop_result()

    def _check_loop_control(self) -> None:
        """Check loop_control configuration for potential problems."""
        task_ctx = TaskContext.current()

        loop_variables = dict(
            loop_var=self._task.loop_control.loop_var,
            index_var=self._task.loop_control.index_var,
        )

        # These reserved variables are set in TaskContext.start_loop,
        # so shouldn't be expected to already be in task variables.
        reserved_loop_variables = {
            "ansible_index_var",
            "ansible_loop",
            "ansible_loop_var",
        }

        duplicate_loop_variables = {
            key for key, group in itertools.groupby(sorted(value for value in loop_variables.values() if value)) if len(list(group)) > 1
        }

        for var_name, var_value in loop_variables.items():
            if not var_value:
                continue

            if var_value in task_ctx.task_vars:
                conflict = "already in use"
            elif var_value in reserved_loop_variables:
                conflict = "reserved"
            elif var_value in C.COMMON_CONNECTION_VARS:
                conflict = "reserved"
            elif var_value in duplicate_loop_variables:
                conflict = "used more than once"
            else:
                continue

            display.warning(
                msg=f"The variable {var_value!r} is {conflict}.",
                help_text=f"You should set the `{var_name}` value in the `loop_control` option for the task "
                          "to something else to avoid variable collisions and unexpected behavior.",
                obj=var_value,
            )

    def _calculate_delegate_to(self):
        """This method is responsible for effectively pre-validating Task.delegate_to and will
        happen before Task.post_validate is executed
        """
        task_ctx = TaskContext.current()

        delegated_vars, delegated_host_name = self._variable_manager.get_delegated_vars_and_hostname(task_ctx.task_templar, self._task, task_ctx.task_vars)
        # At the point this is executed it is safe to mutate self._task,
        # since `self._task` is either a copy referred to by `tmp_task` in `_run_loop`
        # or just a singular non-looped task

        self._task.delegate_to = delegated_host_name  # always override, since a templated result could be an omit (-> None)

        task_ctx.task_vars.update(delegated_vars)

    def _execute(self) -> UnifiedTaskResult:
        utr: UnifiedTaskResult

        task_ctx = TaskContext.current()

        with _display_utils.DeferredWarningContext(variables=task_ctx.task_vars) as warning_ctx:
            try:
                # DTFIX-FUTURE: improve error handling to prioritize the earliest exception, turning the remaining ones into warnings
                utr = self._execute_internal()
                utr.maybe_raise_on_result()
            except (Exception, _task_timeout.TaskTimeoutError) as ex:  # TaskTimeoutError is BaseException
                try:
                    raise AnsibleTaskError(obj=self._task.get_ds()) from ex
                except AnsibleTaskError as atex:
                    utr = UnifiedTaskResult.create_from_action_exception(atex, accept_result_contribution=True)

            self._task.update_result_no_log(task_ctx.task_templar, utr)

        # The warnings/deprecations in the result have already been captured in the DeferredWarningContext by _apply_task_result_compat.
        # The captured warnings/deprecations are a superset of the ones from the result, and may have been converted from a dict to a dataclass.
        # These are then used to supersede the entries in the result.

        utr.finalize_warnings(warning_ctx)

        return utr

    def _update_task_connection(self, task: Task | None = None) -> None:
        """If a connection plugin is loaded, ensure the resolved name is propagated back to the controller as the task's connection."""
        if not task:
            task = self._task

        # FUTURE: What value should be reported when there is no connection?
        #         This is currently not possible, but it should be.

        if isinstance(self._connection, ConnectionBase):
            task.connection = self._connection.ansible_name

    def _execute_internal(self) -> UnifiedTaskResult:
        """
        The primary workhorse of the executor system, this runs the task
        on the specified host (which may be the delegated_to host) and handles
        the retry/until and block rescue/always execution
        """
        task_ctx = TaskContext.current()

        self._calculate_delegate_to()

        context_validation_error = None

        task_vars_with_magic_vars = task_ctx.task_vars.copy()  # copy of task vars with connection vars erased/augmented for delegate_to

        try:
            # TODO: remove play_context as this does not take delegation nor loops correctly into account,
            # the task itself should hold the correct values for connection/shell/become/terminal plugin options to finalize.
            #  Kept for now for backwards compatibility and a few functions that are still exclusive to it.

            # apply the given task's information to the connection info,
            # which may override some fields already set by the play or
            # the options specified on the command line
            self._play_context = self._play_context.set_task_and_variable_override(task=self._task, variables=task_ctx.task_vars, templar=task_ctx.task_templar)

            # fields set from the play/task may be based on task vars, so we have to
            # do the same kind of post validation step on it here before we use it.
            self._play_context.post_validate(templar=task_ctx.task_templar)

            # now that the play context is finalized, if the remote_addr is not set
            # default to using the host's address field as the remote address
            if not self._play_context.remote_addr:
                self._play_context.remote_addr = self._host.address

            # We also add "magic" variables back into the tempvars dict to make sure
            self._play_context.update_vars(task_vars_with_magic_vars)

        except AnsibleError as e:
            # save the error, which we'll raise later if we don't end up
            # skipping this task during the conditional evaluation step
            context_validation_error = e

        # Evaluate the conditional (if any) for this task, which we do before running
        # the final task post-validation. We do this before the post validation due to
        # the fact that the conditional may specify that the task be skipped due to a
        # variable not being present which would otherwise cause validation to fail
        try:
            conditional_result, conditional_item = self._task._resolve_conditional_with_item(self._task.when, task_vars_with_magic_vars)

            if not conditional_result:
                return UnifiedTaskResult.record_conditional_false(conditional_item)
        except AnsibleError as e:
            # FUTURE: this error handling seems problematic; shouldn't a failed loop expression always be an error, rather than letting an item-oriented `when`
            #  expression be treated as a task-oriented one and sweeping the failure under the rug if it happens to be False? If so, this re-raise and the
            #  one just below it should be removed in favor of letting it fly from get_loop_items.
            try:
                # The original error's obj should contain the failed expression, which should be origin tagged.
                # This error is intended to provide additional context (the message) for the original error.
                # By repeating the original error's obj here, the error display will group the two errors together.
                raise AnsibleError("A 'when' expression failed.", obj=e.obj) from e
            except AnsibleError as ae:
                if self._loop_eval_error is None:
                    raise

                # Display the error from the conditional as well to prevent
                # losing information useful for debugging.
                display.error(ae)

            raise self._loop_eval_error

        # Not skipping, if we had loop error raised earlier we need to raise it now to halt the execution of this task
        if self._loop_eval_error is not None:
            raise self._loop_eval_error

        # if we ran into an error while setting up the PlayContext, raise it now, unless is known issue with delegation
        # and undefined vars (correct values are in cvars later on and connection plugins, if still error, blows up there)

        # DTFIX-FUTURE: this should probably be declaratively handled in post_validate (or better, get rid of play_context)
        if context_validation_error is not None:
            raiseit = True
            if self._task.delegate_to:
                if isinstance(context_validation_error, AnsibleParserError):
                    # parser error, might be cause by undef too
                    if isinstance(context_validation_error.__cause__, AnsibleUndefinedVariable):
                        raiseit = False
                elif isinstance(context_validation_error, AnsibleUndefinedVariable):
                    # DTFIX-FUTURE: should not be possible to hit this now (all are AnsibleFieldAttributeError)?
                    raiseit = False
            if raiseit:
                raise context_validation_error  # pylint: disable=raising-bad-type

        # Now we do final validation on the task, which sets all fields to their final values.
        self._task.post_validate(templar=task_ctx.task_templar.extend(variables=task_vars_with_magic_vars))  # should be handled by a context!

        # if this task is a TaskInclude, we just return now with a success code so the
        # main thread can expand the task list for the given host
        if self._task.action in C._ACTION_INCLUDE_TASKS:
            include_args = self._task.args.copy()
            include_file = include_args.pop('_raw_params', None)

            with UnifiedTaskResult.create_and_record() as utr:
                utr.include_file = include_file
                utr.include_args = include_args
                return utr

        # if this task is a IncludeRole, we just return now with a success code so the main thread can expand the task list for the given host
        elif self._task.action in C._ACTION_INCLUDE_ROLE:
            include_args = self._task.args.copy()

            with UnifiedTaskResult.create_and_record() as utr:
                utr.include_args = include_args
                return utr

        # setup cvars copy, used for all connection related templating
        if self._task.delegate_to:
            # use vars from delegated host (which already include task vars) instead of original host
            cvars = task_ctx.task_vars.get('ansible_delegated_vars', {}).get(self._task.delegate_to, {})
        else:
            # just use normal host vars
            cvars = task_ctx.task_vars

        connection_templar = task_ctx.task_templar.extend(variables=cvars)  # should be managed by a context!

        # use magic var if it exists, if not, let task inheritance do it's thing.
        if cvars.get('ansible_connection') is not None:
            current_connection = connection_templar.template(cvars['ansible_connection'])
        else:
            current_connection = self._task.connection

        # get the connection and the handler for this execution
        if (not self._connection or
                not getattr(self._connection, 'connected', False) or
                not self._connection.matches_name([current_connection]) or
                # pc compare, left here for old plugins, but should be irrelevant for those
                # using get_option, since they are cleared each iteration.
                self._play_context.remote_addr != self._connection._play_context.remote_addr):
            self._connection = self._get_connection(cvars, connection_templar, current_connection)
        else:
            # if connection is reused, its _play_context is no longer valid and needs
            # to be replaced with the one templated above, in case other data changed
            self._connection._play_context = self._play_context
            self._set_become_plugin(cvars, connection_templar, self._connection)

        plugin_vars = self._set_connection_options(cvars, connection_templar)

        # update with connection info (i.e ansible_host/ansible_user)
        self._connection.update_vars(task_ctx.task_vars)

        # TODO: eventually remove as pc is taken out of the resolution path
        # feed back into pc to ensure plugins not using get_option can get correct value
        self._connection._play_context = self._play_context.set_task_and_variable_override(
            task=self._task,
            variables=task_ctx.task_vars,
            templar=task_ctx.task_templar,
        )

        # TODO: eventually remove this block as this should be a 'consequence' of 'forced_local' modules, right now rely on remote_is_local connection
        # special handling for python interpreter for network_os, default to ansible python unless overridden
        if 'ansible_python_interpreter' not in cvars and 'ansible_network_os' in cvars and getattr(self._connection, '_remote_is_local', False):
            # this also avoids 'python discovery'
            cvars['ansible_python_interpreter'] = sys.executable

        # get handler
        self._handler, _module_context = self._get_action_handler_with_module_context(templar=task_ctx.task_templar)

        # self._connection should have its final value for this task/loop-item by this point; record on the task object
        self._update_task_connection()

        retries = 1  # includes the default actual run + retries set by user/default
        if self._task.retries is not None:
            retries += max(0, self._task.retries)
        elif self._task.until:
            retries += 3  # the default is not set in FA because we need to differentiate "unset" value

        delay = self._task.delay
        if delay < 0:
            delay = 1

        display.debug("starting attempt loop")
        for attempt in range(1, retries + 1):
            display.debug("running the handler")
            try:
                # FUTURE: exceptions raised anywhere here bypass failed_when, `until` retries, intra-loop register/register-projections
                with _task_timeout.TaskTimeoutError.alarm_timeout(self._task.timeout):
                    try:
                        task_ctx.pending_changes = _task.PendingChanges()

                        with UnifiedTaskResult.create_and_record(self._handler.run(task_vars=task_ctx.task_vars)) as utr:
                            utr.pending_changes = task_ctx.pending_changes
                    finally:
                        task_ctx.pending_changes = None

            finally:
                self._handler.cleanup()
            display.debug("handler run complete")

            if self._task.async_val > 0:
                if self._task.poll > 0 and not utr.skipped and not utr.failed:
                    utr = self._poll_async_result(utr=utr, templar=task_ctx.task_templar, task_vars=task_ctx.task_vars)

                    if utr.failed:
                        self._final_q.send_callback('v2_runner_on_async_failed', self._host, self._task, utr)
                    else:
                        self._final_q.send_callback('v2_runner_on_async_ok', self._host, self._task, utr)

            if utr.ansible_facts and _task.VariableLayer.CACHEABLE_FACT not in utr.pending_changes.register_host_variables:
                # For backward compatibility, if the action provided ansible_facts, use that as the CACHEABLE_FACT layer if the action did not provide one.
                utr.pending_changes.register_host_variables[_task.VariableLayer.CACHEABLE_FACT] = utr.ansible_facts

            # Variable layers should be reflected on task vars in the same way they will be handled by variable manager.
            # What occurs below is a partial re-implementation of variable manager, and thus does not fully reflect its behavior.
            # These updates are only done for the current host.

            if not self._task.delegate_to or not self._task.delegate_facts:
                if cacheable_fact_layer := utr.pending_changes.register_host_variables.get(_task.VariableLayer.CACHEABLE_FACT):
                    task_ctx.update_task_vars(dict(
                        ansible_facts=combine_vars(
                            task_ctx.task_vars.get('ansible_facts', {}),
                            namespace_facts(cacheable_fact_layer)['ansible_facts'],
                        ),
                    ))

                    if _INJECT_FACTS:
                        task_ctx.update_task_vars(_clean_and_deprecate_top_level_facts(cacheable_fact_layer))

                if include_vars_layer := utr.pending_changes.register_host_variables.get(_task.VariableLayer.INCLUDE_VARS):
                    task_ctx.update_task_vars(include_vars_layer)

                if ephemeral_fact_layer := utr.pending_changes.register_host_variables.get(_task.VariableLayer.EPHEMERAL_FACT):
                    task_ctx.update_task_vars(ephemeral_fact_layer)

            if register_vars_layer := utr.pending_changes.register_host_variables.get(_task.VariableLayer.REGISTER_VARS):
                task_ctx.update_task_vars(register_vars_layer)

            # Make attempts and retries available early to allow their use in changed/failed_when
            if retries > 1:
                utr.attempts = attempt

            # if we didn't skip this task, use the helpers to evaluate the changed/
            # failed_when properties
            if not utr.skipped:
                try:
                    if self._task.changed_when:
                        utr.set_changed_when_result(self._task._resolve_conditional(self._task.changed_when, task_ctx.task_vars))
                except AnsibleError as e:
                    utr.set_changed_when_result(e)
                else:
                    try:
                        if self._task.failed_when:
                            utr.set_failed_when_result(self._task._resolve_conditional(self._task.failed_when, task_ctx.task_vars))
                    except AnsibleError as e:
                        utr.set_failed_when_result(e)

            if retries > 1:
                try:
                    if self._task._resolve_conditional(self._task.until or [not utr.failed], task_ctx.task_vars):
                        break
                except AnsibleError as e:
                    # The original error's obj should contain the failed expression, which should be origin tagged.
                    # This error is intended to provide additional context (the message) for the original error.
                    # By repeating the original error's obj here, the error display will group the two errors together.
                    raise AnsibleError("An 'until' expression failed.", obj=e.obj) from e

                # no conditional check, or it failed, so sleep for the specified time
                if attempt < retries:
                    utr.retries = retries
                    utr.attempts = attempt + 1
                    display.debug('Retrying task, attempt %d of %d' % (attempt, retries))
                    self._final_q.send_callback('v2_runner_retry', self._host, self._task, utr)
                    time.sleep(delay)
                    self._handler = self._get_action_handler(templar=task_ctx.task_templar)
        else:
            if retries > 1:
                # we ran out of attempts, so mark the result as failed
                utr.attempts = retries - 1
                utr.failed = True

        # save the notification target in the result, if it was specified, as
        # this task may be running in a loop in which case the notification
        # may be item-specific, ie. "notify: service {{item}}"
        if self._task.notify is not None:
            utr.notify = self._task.notify

        # add the delegated vars to the result, so we can reference them
        # on the results side without having to do any further templating
        # also now add connection vars results when delegating
        if self._task.delegate_to:
            utr.delegated_host = self._task.delegate_to
            utr.callback_delegated_vars_subset = dict(
                ansible_delegated_host=self._task.delegate_to,
                ansible_connection=current_connection,
            )

            # note: here for callbacks that rely on this info to display delegation
            for plugin_var_name in plugin_vars:
                if plugin_var_name not in _DELEGATED_CONNECTION_PLUGIN_VAR_NAMES:
                    continue

                # FUTURE: this is horribly inefficient
                for plugin_option_name in C.config.get_plugin_options_from_var("connection", current_connection, plugin_var_name):
                    utr.callback_delegated_vars_subset[plugin_var_name] = self._connection.get_option(plugin_option_name)

        # and return
        display.debug("attempt loop complete, returning result")
        return utr

    def _poll_async_result(self, utr: UnifiedTaskResult, templar: TemplateEngine, task_vars: dict[str, t.Any]) -> UnifiedTaskResult:
        """
        Polls for the specified JID to be complete
        """
        async_jid = utr.async_job_id

        if async_jid is None:
            # RPFIX-9: FUTURE: why not raise?
            with UnifiedTaskResult.create_and_record() as utr:
                utr.failed = True
                utr.msg = "No job id was returned by the async task"

            return utr

        # Create a new pseudo-task to run the async_status module, and run
        # that (with a sleep for "poll" seconds between each retry) until the
        # async time limit is exceeded.

        async_task = Task.load(dict(
            action='async_status',
            args={'jid': async_jid},
            check_mode=self._task.check_mode,
            environment=self._task.environment,
            delegate_to=self._task.delegate_to,
        ))

        # ensure that the synthetic async task has the resolved connection recorded on it
        self._update_task_connection(async_task)

        # FIXME: this is no longer the case, normal takes care of all, see if this can just be generalized
        # Because this is an async task, the action handler is async. However,
        # we need the 'normal' action handler for the status check, so get it
        # now via the action_loader
        async_handler: ActionBase = self._shared_loader_obj.action_loader.get(
            'ansible.legacy.async_status',
            task=async_task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=Templar._from_template_engine(templar),
            shared_loader_obj=self._shared_loader_obj,
        )

        time_left = self._task.async_val
        while time_left > 0:
            time.sleep(self._task.poll)

            try:
                with UnifiedTaskResult.create_and_record(async_handler.run(task_vars=task_vars)) as async_utr:
                    pass
                # We do not bail out of the loop in cases where the failure
                # is associated with a parsing error. The async_runner can
                # have issues which result in a half-written/unparsable result
                # file on disk, which manifests to the user as a timeout happening
                # before it's time to timeout.
                if async_utr.finished or (async_utr.failed and async_utr.ansible_parsed) or async_utr.skipped:
                    break
            except Exception as e:
                # Connections can raise exceptions during polling (eg, network bounce, reboot); these should be non-fatal.
                # On an exception, call the connection's reset method if it has one
                # (eg, drop/recreate WinRM connection; some reused connections are in a broken state)
                display.vvvv("Exception during async poll, retrying... (%s)" % to_text(e))
                display.debug("Async poll exception was:\n%s" % to_text(traceback.format_exc()))
                try:
                    async_handler._connection.reset()
                except AttributeError:
                    pass

                # Little hack to raise the exception if we've exhausted the timeout period
                time_left -= self._task.poll
                if time_left <= 0:
                    raise
            else:
                time_left -= self._task.poll
                self._final_q.send_callback('v2_runner_on_async_poll', self._host, async_task, async_utr)

        if not async_utr.finished:
            # RPFIX-9: FUTURE: why not raise?
            with UnifiedTaskResult.create_and_record() as async_failed_utr:
                async_failed_utr.failed = True
                async_failed_utr.async_result = async_utr.as_result_dict()

                if async_failed_utr.ansible_parsed:
                    async_failed_utr.msg = "async task did not complete within the requested time - %ss" % self._task.async_val
                else:
                    async_failed_utr.msg = "async task produced unparsable results"

            async_utr = async_failed_utr  # replace the actual async result with a synthesized failure
        else:
            # If the async task finished, automatically cleanup the temporary
            # status file left behind.
            cleanup_task = Task.load(
                {
                    'async_status': {
                        'jid': async_jid,
                        'mode': 'cleanup',
                    },
                    'check_mode': self._task.check_mode,
                    'environment': self._task.environment,
                    'delegate_to': self._task.delegate_to,
                }
            )
            cleanup_handler: ActionBase = self._shared_loader_obj.action_loader.get(
                'ansible.legacy.async_status',
                task=cleanup_task,
                connection=self._connection,
                play_context=self._play_context,
                loader=self._loader,
                templar=Templar._from_template_engine(templar),
                shared_loader_obj=self._shared_loader_obj,
            )
            cleanup_handler.run(task_vars=task_vars)
            cleanup_handler.cleanup(force=True)
            async_handler.cleanup(force=True)

        return async_utr

    def _get_become(self, name):
        become = become_loader.get(name)
        if not become:
            raise AnsibleError("Invalid become method specified, could not find matching plugin: '%s'. "
                               "Use `ansible-doc -t become -l` to list available plugins." % name)
        return become

    def _get_connection(self, cvars, templar, current_connection):
        """
        Reads the connection property for the host, and returns the
        correct connection object from the list of connection plugins
        """

        self._play_context.connection = current_connection

        conn_type = self._play_context.connection

        connection, plugin_load_context = self._shared_loader_obj.connection_loader.get_with_context(
            conn_type,
            self._play_context,
            new_stdin=None,  # No longer used, kept for backwards compat for plugins that explicitly accept this as an arg
            task_uuid=self._task._uuid,
            ansible_playbook_pid=to_text(os.getppid())
        )

        if not connection:
            raise AnsibleError("the connection plugin '%s' was not found" % conn_type)

        self._set_become_plugin(cvars, templar, connection)

        # Also backwards compat call for those still using play_context
        self._play_context.set_attributes_from_plugin(connection)

        return connection

    def _set_become_plugin(self, cvars, templar, connection):
        # load become plugin if needed
        if cvars.get('ansible_become') is not None:
            become = boolean(templar.template(cvars['ansible_become']))
        else:
            become = self._task.become

        if become:
            if cvars.get('ansible_become_method'):
                become_plugin = self._get_become(templar.template(cvars['ansible_become_method']))
            else:
                become_plugin = self._get_become(self._task.become_method)

        else:
            # If become is not enabled on the task it needs to be removed from the connection plugin
            # https://github.com/ansible/ansible/issues/78425
            become_plugin = None

        try:
            connection.set_become_plugin(become_plugin)
        except AttributeError:
            # Older connection plugin that does not support set_become_plugin
            pass

        if become_plugin:
            if getattr(connection.become, 'require_tty', False) and not getattr(connection, 'has_tty', False):
                raise AnsibleError(
                    "The '%s' connection does not provide a TTY which is required for the selected "
                    "become plugin: %s." % (connection._load_name, become_plugin.name)
                )

            # Backwards compat for connection plugins that don't support become plugins
            # Just do this unconditionally for now, we could move it inside of the
            # AttributeError above later
            self._play_context.set_become_plugin(become_plugin.name)

    def _set_plugin_options(self, plugin_type, variables, templar, task_keys):
        try:
            plugin = getattr(self._connection, '_%s' % plugin_type)
        except AttributeError:
            # Some plugins are assigned to private attrs, ``become`` is not
            plugin = getattr(self._connection, plugin_type)

        # network_cli's "real" connection plugin is not named connection
        # to avoid the confusion of having connection.connection
        if plugin_type == "ssh_type_conn":
            plugin_type = "connection"
        option_vars = C.config.get_plugin_vars(plugin_type, plugin._load_name)
        options = {}
        for k in option_vars:
            if k in variables:
                try:
                    options[k] = templar.template(variables[k])
                except AnsibleValueOmittedError:
                    pass

        # TODO move to task method?
        plugin.set_options(task_keys=task_keys, var_options=options)

        return option_vars

    def _set_connection_options(self, variables, templar):

        # keep list of variable names possibly consumed
        varnames = []

        # grab list of usable vars for this plugin
        option_vars = C.config.get_plugin_vars('connection', self._connection._load_name)
        varnames.extend(option_vars)

        task_keys = self._task.dump_attrs()

        # The task_keys 'timeout' attr is the task's timeout, not the connection timeout.
        # The connection timeout is threaded through the play_context for now.
        task_keys['timeout'] = self._play_context.timeout

        if self._play_context.password:
            # The connection password is threaded through the play_context for
            # now. This is something we ultimately want to avoid, but the first
            # step is to get connection plugins pulling the password through the
            # config system instead of directly accessing play_context.
            task_keys['password'] = self._play_context.password

        # Prevent task retries from overriding connection retries
        del task_keys['retries']

        # set options with 'templated vars' specific to this plugin and dependent ones
        var_options = self._connection._resolve_option_variables(variables, templar)
        self._connection.set_options(task_keys=task_keys, var_options=var_options)
        varnames.extend(self._set_plugin_options('shell', variables, templar, task_keys))

        if self._connection.become is not None:
            if self._play_context.become_pass:
                # FIXME: eventually remove from task and play_context, here for backwards compat
                # keep out of play objects to avoid accidental disclosure, only become plugin should have
                # The become pass is already in the play_context if given on
                # the CLI (-K). Make the plugin aware of it in this case.
                task_keys['become_pass'] = self._play_context.become_pass

            varnames.extend(self._set_plugin_options('become', variables, templar, task_keys))

            # FOR BACKWARDS COMPAT:
            for option in ('become_user', 'become_flags', 'become_exe', 'become_pass'):
                try:
                    setattr(self._play_context, option, self._connection.become.get_option(option))
                except KeyError:
                    pass  # some plugins don't support all base flags
            self._play_context.prompt = self._connection.become.prompt

        # deals with networking sub_plugins (network_cli/httpapi/netconf)
        sub = getattr(self._connection, '_sub_plugin', None)
        if sub and sub.get('type') != 'external':
            plugin_type = get_plugin_class(sub.get("obj"))
            varnames.extend(self._set_plugin_options(plugin_type, variables, templar, task_keys))
        sub_conn = getattr(self._connection, 'ssh_type_conn', None)
        if sub_conn is not None:
            varnames.extend(self._set_plugin_options("ssh_type_conn", variables, templar, task_keys))

        return varnames

    def _get_action_handler(self, templar: TemplateEngine) -> ActionBase:
        """
        Returns the correct action plugin to handle the requestion task action
        """
        return self._get_action_handler_with_module_context(templar)[0]

    def _get_action_handler_with_module_context(self, templar: TemplateEngine) -> tuple[ActionBase, PluginLoadContext]:
        """
        Returns the correct action plugin to handle the requestion task action and the module context
        """
        module_collection, separator, module_name = self._task.action.rpartition(".")
        module_prefix = module_name.split('_')[0]
        if module_collection:
            # For network modules, which look for one action plugin per platform, look for the
            # action plugin in the same collection as the module by prefixing the action plugin
            # with the same collection.
            network_action = "{0}.{1}".format(module_collection, module_prefix)
        else:
            network_action = module_prefix

        collections = self._task.collections

        # Check if the module has specified an action handler
        module = self._shared_loader_obj.module_loader.find_plugin_with_context(
            self._task.action, collection_list=collections
        )
        if not module.resolved or not module.action_plugin:
            module = None
        if module is not None:
            handler_name = module.action_plugin
        # let action plugin override module, fallback to 'normal' action plugin otherwise
        elif self._shared_loader_obj.action_loader.has_plugin(self._task.action, collection_list=collections):
            handler_name = self._task.action
        elif module_prefix in C.NETWORK_GROUP_MODULES and self._shared_loader_obj.action_loader.has_plugin(network_action, collection_list=collections):
            handler_name = network_action
            display.vvvv("Using network group action {handler} for {action}".format(handler=handler_name,
                                                                                    action=self._task.action),
                         host=self._play_context.remote_addr)
        else:
            # use ansible.legacy.normal to allow (historic) local action_plugins/ override without collections search
            handler_name = 'ansible.legacy.normal'
            collections = None  # until then, we don't want the task's collection list to be consulted; use the builtin

        # networking/psersistent connections handling
        if any(((self._connection.supports_persistence and C.USE_PERSISTENT_CONNECTIONS), self._connection.force_persistence)):

            # check handler in case we dont need to do all the work to setup persistent connection
            handler_class = self._shared_loader_obj.action_loader.get(handler_name, class_only=True)
            if getattr(handler_class, '_requires_connection', True):
                # for persistent connections, initialize socket path and start connection manager
                self._play_context.timeout = self._connection.get_option('persistent_command_timeout')
                display.vvvv('attempting to start connection', host=self._play_context.remote_addr)
                display.vvvv('using connection plugin %s' % self._connection.transport, host=self._play_context.remote_addr)

                options = self._connection.get_options()
                socket_path = start_connection(self._play_context, options, self._task._uuid)
                display.vvvv('local domain socket path is %s' % socket_path, host=self._play_context.remote_addr)
                setattr(self._connection, '_socket_path', socket_path)
            else:
                # TODO: set self._connection to dummy/noop connection, using local for now
                self._connection = self._get_connection({}, templar, 'local')

        handler = self._shared_loader_obj.action_loader.get(
            handler_name,
            task=self._task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=Templar._from_template_engine(templar),
            shared_loader_obj=self._shared_loader_obj,
            collection_list=collections
        )

        if not handler:
            raise AnsibleError("the handler '%s' was not found" % handler_name)

        return handler, module


CLI_STUB_NAME = 'ansible_connection_cli_stub.py'


def start_connection(play_context, options, task_uuid):
    """
    Starts the persistent connection
    """

    env = os.environ.copy()
    env.update({
        # HACK; most of these paths may change during the controller's lifetime
        # (eg, due to late dynamic role includes, multi-playbook execution), without a way
        # to invalidate/update, the persistent connection helper won't always see the same plugins the controller
        # can.
        'ANSIBLE_BECOME_PLUGINS': become_loader.print_paths(),
        'ANSIBLE_CLICONF_PLUGINS': cliconf_loader.print_paths(),
        'ANSIBLE_COLLECTIONS_PATH': to_native(os.pathsep.join(AnsibleCollectionConfig.collection_paths)),
        'ANSIBLE_CONNECTION_PLUGINS': connection_loader.print_paths(),
        'ANSIBLE_HTTPAPI_PLUGINS': httpapi_loader.print_paths(),
        'ANSIBLE_NETCONF_PLUGINS': netconf_loader.print_paths(),
        'ANSIBLE_TERMINAL_PLUGINS': terminal_loader.print_paths(),
    })
    verbosity = []
    if display.verbosity:
        verbosity.append('-%s' % ('v' * display.verbosity))

    if not (cli_stub_path := C.config.get_config_value('_ANSIBLE_CONNECTION_PATH')):
        cli_stub_path = str(pathlib.Path(scripts.__file__).parent / CLI_STUB_NAME)

    p = subprocess.Popen(
        [sys.executable, cli_stub_path, *verbosity, to_text(os.getppid()), to_text(task_uuid)],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env,
    )

    write_to_stream(p.stdin, options)
    write_to_stream(p.stdin, play_context.dump_attrs())

    (stdout, stderr) = p.communicate()

    if p.returncode == 0:
        result = json.loads(to_text(stdout, errors='surrogate_then_replace'))
    else:
        try:
            result = json.loads(to_text(stderr, errors='surrogate_then_replace'))
        except json.decoder.JSONDecodeError:
            result = {'error': to_text(stderr, errors='surrogate_then_replace')}

    if 'messages' in result:
        for level, message in result['messages']:
            if level == 'log':
                display.display(message, log_only=True)
            elif level in ('debug', 'v', 'vv', 'vvv', 'vvvv', 'vvvvv', 'vvvvvv'):
                getattr(display, level)(message, host=play_context.remote_addr)
            else:
                if hasattr(display, level):
                    getattr(display, level)(message)
                else:
                    display.vvvv(message, host=play_context.remote_addr)

    if 'error' in result:
        if display.verbosity > 2:
            if result.get('exception'):
                msg = "The full traceback is:\n" + result['exception']
                display.display(msg, color=C.COLOR_ERROR)
        raise AnsibleError(result['error'])

    return result['socket_path']



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import cmd
import functools
import os
import pprint
import queue
import sys
import threading
import time
import typing as t
import collections.abc as _c

from collections import deque

from ansible import constants as C, constants
from ansible import context
from ansible.errors import AnsibleError, AnsibleTemplateError
from ansible.executor.play_iterator import IteratingStates, PlayIterator
from ansible.executor.process.worker import WorkerProcess
from ansible._internal._task import WireTaskResult, HostTaskResult
from ansible.executor.task_queue_manager import CallbackSend, DisplaySend, PromptSend, TaskQueueManager
from ansible.module_utils.common.text.converters import to_text
from ansible.module_utils.connection import Connection, ConnectionError
from ansible.playbook.handler import Handler
from ansible.playbook.helpers import load_list_of_blocks
from ansible.playbook.included_file import IncludedFile
from ansible.playbook.play_context import PlayContext
from ansible.playbook.task import Task
from ansible.playbook.task_include import TaskInclude
from ansible.plugins import loader as plugin_loader
from ansible._internal._templating._engine import TemplateEngine
from ansible.utils.display import Display
from ansible.utils.fqcn import add_internal_fqcns
from ansible.utils.sentinel import Sentinel
from ansible._internal import _task, _rpc_host

if t.TYPE_CHECKING:
    from ansible.playbook.role_include import IncludeRole

display = Display()

__all__ = ['StrategyBase']


class StrategySentinel:
    pass


_sentinel = StrategySentinel()

if t.TYPE_CHECKING:
    from ansible.inventory.host import Host


def _get_item_vars(utr: _task.UnifiedTaskResult, task: Task) -> dict[str, object]:
    # RPFIX-9: FUTURE: need a more consistent way to reconstitute/replay loop state controller-side (may be resolved when RPC is implemented)
    item_vars = {}

    if task.loop or task.loop_with:
        if utr.loop_var is not None:
            item_vars[utr.loop_var] = utr.loop_item
        if utr.loop_index_var is not None:
            item_vars[utr.loop_index_var] = utr.loop_index
        if utr.loop_item_label is not None:
            item_vars['_ansible_item_label'] = utr.loop_item_label
        if utr.loop_extended is not None:
            item_vars['ansible_loop'] = utr.loop_extended

    return item_vars


def results_thread_main(strategy: StrategyBase) -> None:
    value: object

    while True:
        try:
            result = strategy._final_q.get()
            if isinstance(result, StrategySentinel):
                break
            elif isinstance(result, DisplaySend):
                dmethod = getattr(display, result.method)
                dmethod(*result.args, **result.kwargs)
            elif isinstance(result, CallbackSend):
                task_result = strategy._convert_wire_task_result_to_host_task_result(result.wire_task_result)
                strategy._tqm.send_callback(result.method_name, task_result)
            elif isinstance(result, WireTaskResult):
                result = strategy._convert_wire_task_result_to_host_task_result(result)
                with strategy._results_lock:
                    strategy._results.append(result)
            elif isinstance(result, PromptSend):
                try:
                    value = display.prompt_until(
                        result.prompt,
                        private=result.private,
                        seconds=result.seconds,
                        complete_input=result.complete_input,
                        interrupt_input=result.interrupt_input,
                    )
                except AnsibleError as e:
                    value = e
                except BaseException as e:
                    # relay unexpected errors so bugs in display are reported and don't cause workers to hang
                    try:
                        raise AnsibleError(f"{e}") from e
                    except AnsibleError as e:
                        value = e
                strategy._workers[result.worker_id].worker_queue.put(value)
            else:
                display.warning('Received an invalid object (%s) in the result queue: %r' % (type(result), result))
        except (OSError, EOFError):
            break
        except queue.Empty:
            pass


def debug_closure(func):
    """Closure to wrap ``StrategyBase._process_pending_results`` and invoke the task debugger"""
    @functools.wraps(func)
    def inner(self, iterator: PlayIterator, one_pass: bool = False, max_passes: int | None = None) -> list[HostTaskResult]:
        status_to_stats_map = (
            ('failed', 'failures'),
            ('unreachable', 'dark'),
            ('changed', 'changed'),
            ('skipped', 'skipped'),
        )

        # We don't know the host yet, copy the previous states, for lookup after we process new results
        prev_host_states = iterator.host_states.copy()

        results: list[HostTaskResult] = func(self, iterator, one_pass=one_pass, max_passes=max_passes)
        _processed_results: list[HostTaskResult] = []

        for result in results:
            task = result.task
            host = result.host
            _queued_task_args = self._queued_task_cache.pop((host.name, task._uuid), None)
            task_vars = _queued_task_args['task_vars']
            play_context = _queued_task_args['play_context']
            # Try to grab the previous host state, if it doesn't exist use get_host_state to generate an empty state
            try:
                prev_host_state = prev_host_states[host.name]
            except KeyError:
                prev_host_state = iterator.get_host_state(host)

            while _needs_debugger(task=result.task, utr=result.utr, globally_enabled=self.debugger_active):
                next_action = NextAction()
                dbg = Debugger(task, host, task_vars, play_context, result, next_action)
                dbg.cmdloop()

                if next_action.result == NextAction.REDO:
                    # rollback host state
                    self._tqm.clear_failed_hosts()
                    if task.run_once and iterator._play.strategy in add_internal_fqcns(('linear',)) and result.utr.failed:
                        for host_name, state in prev_host_states.items():
                            if host_name == host.name:
                                continue
                            iterator.set_state_for_host(host_name, state)
                            iterator._play._removed_hosts.remove(host_name)
                    iterator.set_state_for_host(host.name, prev_host_state)
                    for attr, what in status_to_stats_map:
                        if getattr(result.utr, attr):
                            self._tqm._stats.decrement(what, host.name)
                    self._tqm._stats.decrement('ok', host.name)

                    # redo
                    self._queue_task(host, task, task_vars, play_context)

                    _processed_results.extend(debug_closure(func)(self, iterator, one_pass))
                    break
                elif next_action.result == NextAction.CONTINUE:
                    _processed_results.append(result)
                    break
                elif next_action.result == NextAction.EXIT:
                    # Matches KeyboardInterrupt from bin/ansible
                    sys.exit(99)
            else:
                _processed_results.append(result)

        return _processed_results
    return inner


def _needs_debugger(task: Task, utr: _task.UnifiedTaskResult, globally_enabled: bool = False) -> bool:
    ignore_errors: bool = constants.config.get_config_value('TASK_DEBUGGER_IGNORE_ERRORS') and utr.ignore_errors
    ret = globally_enabled and ((utr.failed and not ignore_errors) or bool(utr.unreachable))

    match task.debugger:
        case 'always':
            ret = True
        case 'never':
            ret = False
        case 'on_failed' if utr.failed and not ignore_errors:
            ret = True
        case 'on_unreachable' if utr.unreachable:
            ret = True
        case 'on_skipped' if utr.skipped:
            ret = True

    return ret


class StrategyBase:

    """
    This is the base class for strategy plugins, which contains some common
    code useful to all strategies like running handlers, cleanup actions, etc.
    """

    # by default, strategies should support throttling but we allow individual
    # strategies to disable this and either forego supporting it or managing
    # the throttling internally (as `free` does)
    ALLOW_BASE_THROTTLING = True

    def __init__(self, tqm: TaskQueueManager) -> None:
        self._tqm = tqm
        self._inventory = tqm.get_inventory()
        self._workers = tqm._workers
        self._variable_manager = tqm.get_variable_manager()
        self._loader = tqm.get_loader()
        self._final_q = tqm._final_q
        self._step = context.CLIARGS.get('step', False)
        self._diff = context.CLIARGS.get('diff', False)
        self._rpc_queue: queue.Queue[_rpc_host.AsyncRPCOperation] = queue.Queue()
        """Pending worker AsyncRPCOperations that need to be serviced under the result thread."""

        # the task cache is a dictionary of tuples of (host.name, task._uuid)
        # used to find the original task object of in-flight tasks and to store
        # the task args/vars and play context info used to queue the task.
        self._queued_task_cache: dict[tuple[str, str], dict[str, t.Any]] = {}

        # Backwards compat: self._display isn't really needed, just import the global display and use that.
        self._display = display

        # internal counters
        self._pending_results = 0
        self._cur_worker = 0

        # this dictionary is used to keep track of hosts that have
        # outstanding tasks still in queue
        self._blocked_hosts: dict[str, bool] = dict()

        self._results: deque[HostTaskResult] = deque()
        self._results_lock = threading.Condition(threading.Lock())

        # create the result processing thread for reading results in the background
        self._results_thread = threading.Thread(target=results_thread_main, args=(self,))
        self._results_thread.daemon = True
        self._results_thread.start()

        # holds the list of active (persistent) connections to be shutdown at
        # play completion
        self._active_connections: dict[Host, str] = dict()

        # Caches for get_host calls, to avoid calling excessively
        # These values should be set at the top of the ``run`` method of each
        # strategy plugin. Use ``_set_hosts_cache`` to set these values
        self._hosts_cache: list[str] = []
        self._hosts_cache_all: list[str] = []

        self.debugger_active = C.ENABLE_TASK_DEBUGGER

    def _set_hosts_cache(self, play, refresh=True):
        """Responsible for setting _hosts_cache and _hosts_cache_all

        See comment in ``__init__`` for the purpose of these caches
        """
        if not refresh and all((self._hosts_cache, self._hosts_cache_all)):
            return

        if not play.finalized and TemplateEngine().is_template(play.hosts):
            _pattern = 'all'
        else:
            _pattern = play.hosts or 'all'
        self._hosts_cache_all = [h.name for h in self._inventory.get_hosts(pattern=_pattern, ignore_restrictions=True)]
        self._hosts_cache = [h.name for h in self._inventory.get_hosts(play.hosts, order=play.order)]

    def cleanup(self):
        # close active persistent connections
        for sock in self._active_connections.values():
            try:
                conn = Connection(sock)
                conn.reset()
            except ConnectionError as e:
                # most likely socket is already closed
                display.debug("got an error while closing persistent connection: %s" % e)
        self._final_q.put(_sentinel)
        self._results_thread.join()

    def run(self, iterator: PlayIterator, play_context: PlayContext, result=0):
        # execute one more pass through the iterator without peeking, to
        # make sure that all of the hosts are advanced to their final task.
        # This should be safe, as everything should be IteratingStates.COMPLETE by
        # this point, though the strategy may not advance the hosts itself.

        for host in self._hosts_cache:
            if host not in self._tqm._unreachable_hosts:
                try:
                    iterator.get_next_task_for_host(self._inventory.hosts[host])
                except KeyError:
                    iterator.get_next_task_for_host(self._inventory.get_host(host))

        # return the appropriate code, depending on the status hosts after the run
        if not isinstance(result, bool) and result != self._tqm.RUN_OK:
            return result
        elif len(self._tqm._unreachable_hosts.keys()) > 0:
            return self._tqm.RUN_UNREACHABLE_HOSTS
        elif len(iterator.get_failed_hosts()) > 0:
            return self._tqm.RUN_FAILED_HOSTS
        else:
            return self._tqm.RUN_OK

    def get_hosts_remaining(self, play):
        self._set_hosts_cache(play, refresh=False)
        ignore = set(self._tqm._failed_hosts).union(self._tqm._unreachable_hosts)
        return [host for host in self._hosts_cache if host not in ignore]

    def get_failed_hosts(self, play):
        self._set_hosts_cache(play, refresh=False)
        return [host for host in self._hosts_cache if host in self._tqm._failed_hosts]

    def add_tqm_variables(self, vars, play):
        """
        Base class method to add extra variables/information to the list of task
        vars sent through the executor engine regarding the task queue manager state.
        """
        vars['ansible_current_hosts'] = self.get_hosts_remaining(play)
        vars['ansible_failed_hosts'] = self.get_failed_hosts(play)

    def _queue_task(self, host, task, task_vars, play_context):
        """ handles queueing the task up to be sent to a worker """

        display.debug("entering _queue_task() for %s/%s" % (host.name, task.action))

        # create a templar and template things we need later for the queuing process
        templar = TemplateEngine(loader=self._loader, variables=task_vars)

        try:
            throttle = int(templar.template(task.throttle))
        except Exception as ex:
            raise AnsibleError("Failed to convert the throttle value to an integer.", obj=task.throttle) from ex

        # and then queue the new task
        try:
            # Determine the "rewind point" of the worker list. This means we start
            # iterating over the list of workers until the end of the list is found.
            # Normally, that is simply the length of the workers list (as determined
            # by the forks or serial setting), however a task/block/play may "throttle"
            # that limit down.
            rewind_point = len(self._workers)
            if throttle > 0 and self.ALLOW_BASE_THROTTLING:
                if task.run_once:
                    display.debug("Ignoring 'throttle' as 'run_once' is also set for '%s'" % task.get_name())
                else:
                    if throttle <= rewind_point:
                        display.debug("task: %s, throttle: %d" % (task.get_name(), throttle))
                        rewind_point = throttle

            queued = False
            starting_worker = self._cur_worker
            while True:
                self._process_rpc_queue()

                if self._cur_worker >= rewind_point:
                    self._cur_worker = 0

                worker_prc = self._workers[self._cur_worker]
                if worker_prc is None or not worker_prc.is_alive():
                    if worker_prc:
                        worker_prc.close()
                    self._queued_task_cache[(host.name, task._uuid)] = {
                        'host': host,
                        'task': task,
                        'task_vars': task_vars,
                        'play_context': play_context
                    }

                    # Pass WorkerProcess its strategy worker number so it can send an identifier along with intra-task requests
                    worker_prc = WorkerProcess(
                        final_q=self._final_q,
                        task_vars=task_vars,
                        host=host,
                        task=task,
                        play_context=play_context,
                        loader=self._loader,
                        variable_manager=self._variable_manager,
                        shared_loader_obj=plugin_loader.get_plugin_loader_namespace(),
                        worker_id=self._cur_worker,
                        cliargs=context.CLIARGS,
                    )
                    self._workers[self._cur_worker] = worker_prc
                    self._tqm.send_callback('v2_runner_on_start', host, task)
                    worker_prc.start()
                    display.debug("worker is %d (out of %d available)" % (self._cur_worker + 1, len(self._workers)))
                    queued = True

                self._cur_worker += 1

                if self._cur_worker >= rewind_point:
                    self._cur_worker = 0

                if queued:
                    break
                elif self._cur_worker == starting_worker:
                    time.sleep(0.0001)

            self._pending_results += 1
        except (EOFError, OSError, AssertionError) as ex:
            # most likely an abort
            display.debug(f"got an error while queuing: {ex}")
            return
        display.debug("exiting _queue_task() for %s/%s" % (host.name, task.action))

    def get_task_hosts(self, iterator, task_host, task):
        if task.run_once:
            host_list = [host for host in self._hosts_cache if host not in self._tqm._unreachable_hosts]
        else:
            host_list = [task_host.name]
        return host_list

    def _convert_wire_task_result_to_host_task_result(self, wire_task_result: WireTaskResult) -> HostTaskResult:
        """Rehydrate a `HostTaskResult` from a serialized `WireTaskResult`."""
        host = self._inventory.get_host(wire_task_result.host_name)
        queue_cache_entry = (host.name, wire_task_result.task_uuid)

        try:
            found_task = self._queued_task_cache[queue_cache_entry]['task']
        except KeyError:
            # This should only happen due to an implicit task created by the
            # TaskExecutor, restrict this behavior to the explicit use case
            # of an implicit async_status task
            if wire_task_result.task_fields.get('action') != 'async_status':
                raise

            task = Task()
        else:
            task = found_task.copy(exclude_parent=True, exclude_tasks=True)
            task._parent = found_task._parent

        task.from_attrs(wire_task_result.task_fields)

        return HostTaskResult(
            host=host,
            task=task,
            utr=wire_task_result.utr,
        )

    def search_handlers_by_notification(self, notification: str, iterator: PlayIterator) -> t.Generator[Handler, None, None]:
        handlers = [h for b in reversed(iterator._play.handlers) for h in b.block]
        # iterate in reversed order since last handler loaded with the same name wins
        for handler in handlers:
            if not handler.name:
                continue

            if not handler.cached_name:
                def variables_factory() -> dict[str, t.Any]:
                    return self._variable_manager.get_vars(
                        play=iterator._play,
                        task=handler,
                        _hosts=self._hosts_cache,
                        _hosts_all=self._hosts_cache_all
                    )

                templar = TemplateEngine(variables_factory=variables_factory)

                try:
                    handler.name = templar.template(handler.name)
                except AnsibleTemplateError as e:
                    # We skip this handler due to the fact that it may be using
                    # a variable in the name that was conditionally included via
                    # set_fact or some other method, and we don't want to error
                    # out unnecessarily
                    if not handler.listen:
                        display.warning(
                            "Handler '%s' is unusable because it has no listen topics and "
                            "the name could not be templated (host-specific variables are "
                            "not supported in handler names). The error: %s" % (handler.name, to_text(e))
                        )
                    continue

                handler.cached_name = True

            # first we check with the full result of get_name(), which may
            # include the role name (if the handler is from a role). If that
            # is not found, we resort to the simple name field, which doesn't
            # have anything extra added to it.
            if notification in {
                handler.name,
                handler.get_name(include_role_fqcn=False),
                handler.get_name(include_role_fqcn=True),
            }:
                yield handler
                break

        seen = set()
        for handler in handlers:
            if notification in handler.listen:
                if handler.name and handler.name in seen:
                    continue
                seen.add(handler.name)
                yield handler

    def _process_rpc_queue(self) -> None:
        try:
            while True:
                rpc_request = self._rpc_queue.get(block=False)
                rpc_request.dispatch()
        except queue.Empty:
            pass

    @debug_closure
    def _process_pending_results(self, iterator: PlayIterator, one_pass: bool = False, max_passes: int | None = None) -> list[HostTaskResult]:
        """
        Reads results off the final queue and takes appropriate action
        based on the result (executing callbacks, updating state, etc.).
        """
        ret_results = []
        cur_pass = 0

        while True:
            self._process_rpc_queue()

            try:
                self._results_lock.acquire()
                task_result = self._results.popleft()
            except IndexError:
                break
            finally:
                self._results_lock.release()

            original_host = task_result.host
            original_task: Task = task_result.task

            # all host status messages contain 2 entries: (msg, task_result)
            role_ran = False

            if task_result.utr.failed:
                role_ran = True
                ignore_errors = task_result.utr.ignore_errors

                if not ignore_errors:
                    # save the current state before failing it for later inspection
                    state_when_failed = iterator.get_state_for_host(original_host.name)
                    display.debug("marking %s as failed" % original_host.name)

                    if original_task.run_once:
                        # if we're using run_once, we have to fail every host here
                        for h in self._inventory.get_hosts(iterator._play.hosts):
                            if h.name not in self._tqm._unreachable_hosts:
                                iterator.mark_host_failed(h)
                    else:
                        iterator.mark_host_failed(original_host)

                    state, dummy = iterator.get_next_task_for_host(original_host, peek=True)

                    if iterator.is_failed(original_host) and state and state.run_state == IteratingStates.COMPLETE:
                        self._tqm._failed_hosts[original_host.name] = True

                    # if we're iterating on the rescue portion of a block then
                    # we save the failed task in a special var for use
                    # within the rescue/always
                    if iterator.is_any_block_rescuing(state_when_failed):
                        self._tqm._stats.increment('rescued', original_host.name)

                        iterator._play._removed_hosts.remove(original_host.name)

                        self._variable_manager.set_nonpersistent_facts(
                            original_host.name,
                            dict(
                                ansible_failed_task=original_task.dump_attrs(),
                                ansible_failed_result=task_result.utr.as_result_dict(),
                            ),
                        )
                    else:
                        self._tqm._stats.increment('failures', original_host.name)
                else:
                    self._tqm._stats.increment('ok', original_host.name)
                    self._tqm._stats.increment('ignored', original_host.name)

                    if task_result.utr.changed:
                        self._tqm._stats.increment('changed', original_host.name)

                self._tqm.send_callback('v2_runner_on_failed', task_result, ignore_errors=ignore_errors)
            elif task_result.utr.unreachable:
                if not task_result.utr.ignore_unreachable:
                    self._tqm._unreachable_hosts[original_host.name] = True

                    iterator._play._removed_hosts.append(original_host.name)

                    self._tqm._stats.increment('dark', original_host.name)
                else:
                    self._tqm._stats.increment('ok', original_host.name)
                    self._tqm._stats.increment('ignored', original_host.name)

                self._tqm.send_callback('v2_runner_on_unreachable', task_result)
            elif task_result.utr.skipped:
                self._tqm._stats.increment('skipped', original_host.name)
                self._tqm.send_callback('v2_runner_on_skipped', task_result)
            else:
                role_ran = True

                result_utrs: list[_task.UnifiedTaskResult]

                if original_task.loop:
                    # this task had a loop, and has more than one result, so
                    # loop over all of them instead of a single result
                    result_utrs = task_result.utr.loop_results
                else:
                    result_utrs = [task_result.utr]

                for result_utr in result_utrs:
                    if result_utr.notify and task_result.utr.changed:
                        # only ensure that notified handlers exist, if so save the notifications for when
                        # handlers are actually flushed so the last defined handlers are executed,
                        # otherwise depending on the setting either error or warn
                        host_state = iterator.get_state_for_host(original_host.name)

                        for notification in result_utr.notify:
                            handler: type[Sentinel] | Handler = Sentinel

                            for handler in self.search_handlers_by_notification(notification, iterator):
                                if host_state.run_state == IteratingStates.HANDLERS:
                                    # we're currently iterating handlers, so we need to expand this now
                                    if handler.notify_host(original_host):
                                        # NOTE even with notifications deduplicated this can still happen in case of handlers being
                                        # notified multiple times using different names, like role name or fqcn
                                        self._tqm.send_callback('v2_playbook_on_notify', handler, original_host)
                                else:
                                    iterator.add_notification(original_host.name, notification)
                                    display.vv(f"Notification for handler {notification} has been saved.")
                                    break

                            if handler is Sentinel:
                                msg = (
                                    f"The requested handler '{notification}' was not found in either the main handlers"
                                    " list nor in the listening handlers list"
                                )

                                if C.ERROR_ON_MISSING_HANDLER:
                                    raise AnsibleError(msg)

                                display.warning(msg)

                    if result_utr.pending_changes.register_host_variables:
                        original_host_list = self.get_task_hosts(iterator, original_host, original_task)

                        if original_task.delegate_to and original_task.delegate_facts:
                            potentially_delegated_host_list = [result_utr.delegated_host or original_task.delegate_to]
                        else:
                            potentially_delegated_host_list = original_host_list

                        # Apply changes from register_host_variables to variable manager in sorted order,
                        # to ensure consistent variable precedence based on layer.
                        for variable_layer, variables in sorted(result_utr.pending_changes.register_host_variables.items()):
                            match variable_layer:
                                case _task.VariableLayer.REGISTER_VARS:
                                    for target_host in original_host_list:
                                        self._variable_manager.set_nonpersistent_facts(target_host, variables)
                                case _task.VariableLayer.CACHEABLE_FACT:
                                    for target_host in potentially_delegated_host_list:
                                        self._variable_manager.set_host_facts(target_host, variables)
                                case _task.VariableLayer.EPHEMERAL_FACT:
                                    for target_host in potentially_delegated_host_list:
                                        self._variable_manager.set_nonpersistent_facts(target_host, variables)
                                case _task.VariableLayer.INCLUDE_VARS:
                                    for target_host in potentially_delegated_host_list:
                                        for var_name, var_value in variables.items():
                                            self._variable_manager.set_host_variable(target_host, var_name, var_value)
                                case _:
                                    raise NotImplementedError(f"Unsupported variable layer: {variable_layer}")

                    if result_utr.stats is not None:
                        stats_data = result_utr.stats['data']

                        if result_utr.stats.get('per_host'):
                            host_list = self.get_task_hosts(iterator, original_host, original_task)
                        else:
                            host_list = [None]

                        aggregate = result_utr.stats.get('aggregate')

                        for myhost in host_list:
                            for k in stats_data.keys():
                                if aggregate:
                                    self._tqm._stats.update_custom_stats(k, stats_data[k], myhost)
                                else:
                                    self._tqm._stats.set_custom_stats(k, stats_data[k], myhost)

                if task_result.utr.diff is not None:
                    # this case is only for non-loops; loop item diff callbacks are dispatched directly by TaskExecutor
                    if self._diff or getattr(original_task, 'diff', False):
                        self._tqm.send_callback('v2_on_file_diff', task_result)

                if not isinstance(original_task, TaskInclude):
                    self._tqm._stats.increment('ok', original_host.name)

                    if task_result.utr.changed:
                        self._tqm._stats.increment('changed', original_host.name)

                # finally, send the ok for this task
                self._tqm.send_callback('v2_runner_on_ok', task_result)

            # register final results
            if task_result.utr.registered_values:
                host_list = self.get_task_hosts(iterator, original_host, original_task)

                for target_host in host_list:
                    self._variable_manager.set_nonpersistent_facts(target_host, task_result.utr.registered_values)

            self._pending_results -= 1

            if original_host.name in self._blocked_hosts:
                del self._blocked_hosts[original_host.name]

            # If this is a role task, mark the parent role as being run (if
            # the task was ok or failed, but not skipped or unreachable)
            if original_task._role is not None and role_ran:  # TODO:  and original_task.action not in C._ACTION_INCLUDE_ROLE:?
                # lookup the role in the role cache to make sure we're dealing
                # with the correct object and mark it as executed
                role_obj = self._get_cached_role(original_task, iterator._play)
                role_obj._had_task_run[original_host.name] = True

            ret_results.append(task_result)

            if one_pass or max_passes is not None and (cur_pass + 1) >= max_passes:
                break

            cur_pass += 1

        return ret_results

    def _wait_on_pending_results(self, iterator):
        """
        Wait for the shared counter to drop to zero, using a short sleep
        between checks to ensure we don't spin lock
        """

        ret_results = []

        display.debug("waiting for pending results...")
        while self._pending_results > 0 and not self._tqm._terminated:

            if self._tqm.has_dead_workers():
                raise AnsibleError("A worker was found in a dead state")

            results = self._process_pending_results(iterator)
            ret_results.extend(results)
            if self._pending_results > 0:
                time.sleep(C.DEFAULT_INTERNAL_POLL_INTERVAL)

        display.debug("no more pending results, returning what we have")

        return ret_results

    def _copy_included_file(self, included_file: IncludedFile) -> TaskInclude | IncludeRole:
        """
        A proven safe and performant way to create a copy of an included file
        """
        ti_copy = included_file._task.copy(exclude_parent=True)
        ti_copy._parent = included_file._task._parent

        temp_vars = ti_copy.vars | included_file._vars

        ti_copy.vars = temp_vars

        return ti_copy

    def _load_included_file(self, included_file: IncludedFile, iterator, is_handler=False):
        """
        Loads an included YAML file of tasks, applying the optional set of variables.

        Raises AnsibleError exception in case of a failure during including a file,
        in such case the caller is responsible for marking the host(s) as failed
        using PlayIterator.mark_host_failed().
        """
        display.debug("loading included file: %s" % included_file._filename)
        data = self._loader.load_from_file(included_file._filename, trusted_as_template=True)
        if data is None:
            return []
        elif not isinstance(data, list):
            raise AnsibleError("included task files must contain a list of tasks", obj=data)

        ti_copy = self._copy_included_file(included_file)

        block_list = load_list_of_blocks(
            data,
            play=iterator._play,
            parent_block=ti_copy.build_parent_block(),
            role=included_file._task._role,
            use_handlers=is_handler,
            loader=self._loader,
            variable_manager=self._variable_manager,
        )

        display.debug("done processing included file")
        return block_list

    def _take_step(self, task, host=None):

        ret = False
        msg = u'Perform task: %s ' % task
        if host:
            msg += u'on %s ' % host
        msg += u'(N)o/(y)es/(c)ontinue: '
        resp = display.prompt(msg)

        if resp.lower() in ['y', 'yes']:
            display.debug("User ran task")
            ret = True
        elif resp.lower() in ['c', 'continue']:
            display.debug("User ran task and canceled step mode")
            self._step = False
            ret = True
        else:
            display.debug("User skipped task")

        display.banner(msg)

        return ret

    def _cond_not_supported_warn(self, task_name):
        display.warning("%s task does not support when conditional" % task_name)

    def _execute_meta(self, task: Task, play_context, iterator, target_host: Host):
        task._resolved_action = 'ansible.builtin.meta'  # _post_validate_args is never called for meta actions, so resolved_action hasn't been set

        # meta tasks store their args in the _raw_params field of args,
        # since they do not use k=v pairs, so get that
        meta_action = task._get_meta()

        def _evaluate_conditional(conditional_host: Host) -> bool:
            if not task.when:
                return True

            # FUTURE: If this when fails it will take down the process.
            return task._resolve_conditional(task.when, self._variable_manager.get_vars(
                play=iterator._play, host=conditional_host, task=task, _hosts=self._hosts_cache, _hosts_all=self._hosts_cache_all))

        skipped = False
        msg = meta_action
        skip_reason = '%s conditional evaluated to False' % meta_action
        if isinstance(task, Handler):
            self._tqm.send_callback('v2_playbook_on_handler_task_start', task)
        else:
            self._tqm.send_callback('v2_playbook_on_task_start', task, is_conditional=False)

        # These don't support "when" conditionals
        if meta_action in ('noop', 'refresh_inventory', 'reset_connection') and task.when:
            self._cond_not_supported_warn(meta_action)

        if meta_action == 'noop':
            msg = "noop"
        elif meta_action == 'flush_handlers':
            if _evaluate_conditional(target_host):
                host_state = iterator.get_state_for_host(target_host.name)
                # actually notify proper handlers based on all notifications up to this point
                for notification in list(host_state.handler_notifications):
                    for handler in self.search_handlers_by_notification(notification, iterator):
                        if handler.notify_host(target_host):
                            # NOTE even with notifications deduplicated this can still happen in case of handlers being
                            # notified multiple times using different names, like role name or fqcn
                            self._tqm.send_callback('v2_playbook_on_notify', handler, target_host)
                    iterator.clear_notification(target_host.name, notification)

                if host_state.run_state == IteratingStates.HANDLERS:
                    raise AnsibleError('flush_handlers cannot be used as a handler')
                if target_host.name not in self._tqm._unreachable_hosts:
                    host_state.pre_flushing_run_state = host_state.run_state
                    host_state.run_state = IteratingStates.HANDLERS
                msg = "triggered running handlers for %s" % target_host.name
            else:
                skipped = True
                skip_reason += ', not running handlers for %s' % target_host.name
        elif meta_action == 'refresh_inventory':
            self._inventory.refresh_inventory()
            self._set_hosts_cache(iterator._play)
            msg = "inventory successfully refreshed"
        elif meta_action == 'clear_facts':
            if _evaluate_conditional(target_host):
                for host in self._inventory.get_hosts(iterator._play.hosts):
                    hostname = host.get_name()
                    self._variable_manager.clear_facts(hostname)
                msg = "facts cleared"
            else:
                skipped = True
                skip_reason += ', not clearing facts and fact cache for %s' % target_host.name
        elif meta_action == 'clear_host_errors':
            if _evaluate_conditional(target_host):
                for host in self._inventory.get_hosts(iterator._play.hosts):
                    self._tqm._failed_hosts.pop(host.name, False)
                    self._tqm._unreachable_hosts.pop(host.name, False)
                    iterator.clear_host_errors(host)
                msg = "cleared host errors"
            else:
                skipped = True
                skip_reason += ', not clearing host error state for %s' % target_host.name
        elif meta_action == 'end_batch':
            if _evaluate_conditional(target_host):
                for host in self._inventory.get_hosts(iterator._play.hosts):
                    if host.name not in self._tqm._unreachable_hosts:
                        iterator.end_host(host.name)
                msg = "ending batch"
            else:
                skipped = True
                skip_reason += ', continuing current batch'
        elif meta_action == 'end_play':
            if _evaluate_conditional(target_host):
                for host in self._inventory.get_hosts(iterator._play.hosts):
                    if host.name not in self._tqm._unreachable_hosts:
                        iterator.end_host(host.name)
                        # end_play is used in PlaybookExecutor/TQM to indicate that
                        # the whole play is supposed to be ended as opposed to just a batch
                        iterator.end_play = True
                msg = "ending play"
            else:
                skipped = True
                skip_reason += ', continuing play'
        elif meta_action == 'end_host':
            if _evaluate_conditional(target_host):
                iterator.end_host(target_host.name)
                msg = "ending play for %s" % target_host.name
            else:
                skipped = True
                skip_reason += ", continuing execution for %s" % target_host.name
                # TODO: Nix msg here? Left for historical reasons, but skip_reason exists now.
                msg = "end_host conditional evaluated to false, continuing execution for %s" % target_host.name
        elif meta_action == 'role_complete':
            if task.implicit:
                role_obj = self._get_cached_role(task, iterator._play)
                if target_host.name in role_obj._had_task_run:
                    role_obj._completed[target_host.name] = True
                    msg = 'role_complete for %s' % target_host.name
        elif meta_action == 'end_role':
            if _evaluate_conditional(target_host):
                while True:
                    state, task = iterator.get_next_task_for_host(target_host, peek=True)
                    if task._get_meta() == "role_complete":
                        break
                    iterator.set_state_for_host(target_host.name, state)
                    display.debug("'%s' skipped because role has been ended via 'end_role'" % task)
                msg = 'ending role %s for %s' % (task._role.get_name(), target_host.name)
            else:
                skipped = True
                skip_reason += ', continuing role %s for %s' % (task._role.get_name(), target_host.name)
        elif meta_action == 'reset_connection':
            all_vars = self._variable_manager.get_vars(play=iterator._play, host=target_host, task=task,
                                                       _hosts=self._hosts_cache, _hosts_all=self._hosts_cache_all)
            templar = TemplateEngine(loader=self._loader, variables=all_vars)

            # apply the given task's information to the connection info,
            # which may override some fields already set by the play or
            # the options specified on the command line
            play_context = play_context.set_task_and_variable_override(task=task, variables=all_vars, templar=templar)

            # fields set from the play/task may be based on variables, so we have to
            # do the same kind of post validation step on it here before we use it.
            play_context.post_validate(templar=templar)

            # now that the play context is finalized, if the remote_addr is not set
            # default to using the host's address field as the remote address
            if not play_context.remote_addr:
                play_context.remote_addr = target_host.address

            # We also add "magic" variables back into the variables dict to make sure
            # a certain subset of variables exist. This 'mostly' works here cause meta
            # disregards the loop, but should not really use play_context at all
            play_context.update_vars(all_vars)

            if target_host in self._active_connections:
                connection = Connection(self._active_connections[target_host])
                del self._active_connections[target_host]
            else:
                connection = plugin_loader.connection_loader.get(play_context.connection, play_context, os.devnull)
                var_options = connection._resolve_option_variables(all_vars, templar)
                connection.set_options(task_keys=task.dump_attrs(), var_options=var_options)
                play_context.set_attributes_from_plugin(connection)

            if connection:
                try:
                    connection.reset()
                    msg = 'reset connection'
                except ConnectionError as e:
                    # most likely socket is already closed
                    display.debug("got an error while closing persistent connection: %s" % e)
            else:
                msg = 'no connection, nothing to reset'
        else:
            raise AnsibleError("invalid meta action requested: %s" % meta_action, obj=task._ds)

        utr = _task.UnifiedTaskResult.from_action_result_dict({})
        utr.msg = msg

        if skipped:
            utr.set_skipped(skip_reason)

        header = skip_reason if skipped else msg
        if task.implicit:
            display.debug(f"META: {header}")
        else:
            display.vv(f"META: {header}")

        htr = HostTaskResult(host=target_host, task=task, utr=utr)

        if skipped:
            self._tqm.send_callback('v2_runner_on_skipped', htr)

        return [htr]

    def _get_cached_role(self, task, play):
        return play._get_cached_role(task._role)

    def get_hosts_left(self, iterator):
        """ returns list of available hosts for this iterator by filtering out unreachables """

        hosts_left = []
        for host in self._hosts_cache:
            if host not in self._tqm._unreachable_hosts:
                try:
                    hosts_left.append(self._inventory.hosts[host])
                except KeyError:
                    hosts_left.append(self._inventory.get_host(host))
        return hosts_left

    def update_active_connections(self, results: _c.Iterable[HostTaskResult]) -> None:
        """ updates the current active persistent connections """
        for result in results:
            if socket_path := result.task.args.get('_ansible_socket'):
                if result.host not in self._active_connections:
                    self._active_connections[result.host] = socket_path


class NextAction(object):
    """ The next action after an interpreter's exit. """
    REDO = 1
    CONTINUE = 2
    EXIT = 3

    def __init__(self, result=EXIT):
        self.result = result


class Debugger(cmd.Cmd):
    prompt_continuous = '> '  # multiple lines

    def __init__(self, task, host, task_vars, play_context, result, next_action):
        # cmd.Cmd is old-style class
        cmd.Cmd.__init__(self)

        self.prompt = '[%s] %s (debug)> ' % (host, task)
        self.intro = None
        self.scope = {}
        self.scope['task'] = task
        self.scope['task_vars'] = task_vars
        self.scope['host'] = host
        self.scope['play_context'] = play_context
        self.scope['result'] = result
        self.next_action = next_action

    def cmdloop(self):
        try:
            cmd.Cmd.cmdloop(self)
        except KeyboardInterrupt:
            pass

    do_h = cmd.Cmd.do_help

    def do_EOF(self, args):
        """Quit"""
        return self.do_quit(args)

    def do_quit(self, args):
        """Quit"""
        display.display('User interrupted execution')
        self.next_action.result = NextAction.EXIT
        return True

    do_q = do_quit

    def do_continue(self, args):
        """Continue to next result"""
        self.next_action.result = NextAction.CONTINUE
        return True

    do_c = do_continue

    def do_redo(self, args):
        """Schedule task for re-execution. The re-execution may not be the next result"""
        self.next_action.result = NextAction.REDO
        return True

    do_r = do_redo

    def do_update_task(self, args):
        """Recreate the task from ``task._ds``, and template with updated ``task_vars``"""
        templar = TemplateEngine(variables=self.scope['task_vars'])
        task = self.scope['task']
        task = task.load_data(task._ds)
        task.post_validate(templar)
        self.scope['task'] = task

    do_u = do_update_task

    def evaluate(self, args):
        try:
            return eval(args, globals(), self.scope)
        except Exception:
            t, v = sys.exc_info()[:2]
            if isinstance(t, str):
                exc_type_name = t
            else:
                exc_type_name = t.__name__
            display.display('***%s:%s' % (exc_type_name, repr(v)))
            raise

    def do_pprint(self, args):
        """Pretty Print"""
        try:
            result = self.evaluate(args)
            display.display(pprint.pformat(result))
        except Exception:
            pass

    do_p = do_pprint

    def execute(self, args):
        try:
            code = compile(args + '\n', '<stdin>', 'single')
            exec(code, globals(), self.scope)
        except Exception:
            t, v = sys.exc_info()[:2]
            if isinstance(t, str):
                exc_type_name = t
            else:
                exc_type_name = t.__name__
            display.display('***%s:%s' % (exc_type_name, repr(v)))
            raise

    def default(self, line):
        try:
            self.execute(line)
        except Exception:
            pass



# -*- coding: utf-8 -*-
# Copyright: (c) 2020-2021, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""Dependency resolution machinery."""

from __future__ import annotations

import collections.abc as _c
import typing as t

if t.TYPE_CHECKING:
    from ansible.galaxy.api import GalaxyAPI
    from ansible.galaxy.collection.concrete_artifact_manager import (
        ConcreteArtifactsManager,
    )
    from ansible.galaxy.dependency_resolution.dataclasses import Candidate

from ansible.galaxy.collection.galaxy_api_proxy import MultiGalaxyAPIProxy
from ansible.galaxy.dependency_resolution.providers import CollectionDependencyProvider
from ansible.galaxy.dependency_resolution.reporters import CollectionDependencyReporter
from ansible.galaxy.dependency_resolution.resolvers import CollectionDependencyResolver


def build_collection_dependency_resolver(
        galaxy_apis: _c.Iterable[GalaxyAPI],
        concrete_artifacts_manager: ConcreteArtifactsManager,
        preferred_candidates: _c.Iterable[Candidate] | None = None,
        with_deps: bool = True,
        with_pre_releases: bool = False,
        upgrade: bool = False,
        include_signatures: bool = True,
        offline: bool = False,
) -> CollectionDependencyResolver:
    """Return a collection dependency resolver.

    The returned instance will have a ``resolve()`` method for
    further consumption.
    """
    return CollectionDependencyResolver(
        CollectionDependencyProvider(
            apis=MultiGalaxyAPIProxy(galaxy_apis, concrete_artifacts_manager, offline=offline),
            concrete_artifacts_manager=concrete_artifacts_manager,
            preferred_candidates=preferred_candidates,
            with_deps=with_deps,
            with_pre_releases=with_pre_releases,
            upgrade=upgrade,
            include_signatures=include_signatures,
        ),
        CollectionDependencyReporter(),
    )



from __future__ import annotations

import contextlib
import dataclasses
import typing as t

if t.TYPE_CHECKING:
    from ansible.plugins.strategy import StrategyBase
    from ansible.executor.task_queue_manager import TaskQueueManager


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class StrategyContext:
    """Expose active strategy and TQM instances."""

    _current: t.ClassVar[t.Self | None] = None

    strategy: StrategyBase
    tqm: TaskQueueManager

    @contextlib.contextmanager
    def activate(self) -> t.Generator[None]:
        cls = type(self)

        try:
            cls._current = self
            yield
        finally:
            cls._current = None

    @classmethod
    def current(cls) -> StrategyContext:
        if not cls._current:
            raise ReferenceError(f"A required {cls.__name__} context is not active.") from None

        return cls._current



"""Ansible callback plugin to enforce an execution deadline using faulthandler."""
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import datetime
import faulthandler
import os
import pathlib
import shlex
import sys

from ansible.plugins.callback import CallbackBase

DOCUMENTATION = """
    name: timeout
    type: aggregate
    short_description: Dump thread stacks when an execution deadline is reached.
    description:
      - Enforces an execution deadline by dumping all thread stacks to a file
        and terminating the process when the deadline expires.
      - Intended for use by ansible-test only. Do not enable manually.
    options:
      deadline:
        description: Execution deadline as a UTC epoch timestamp.
        env:
          - name: ANSIBLE_TEST_TIMEOUT_DEADLINE
      dump_dir:
        description: Directory for timeout dump files.
        env:
          - name: ANSIBLE_TEST_TIMEOUT_DIR
"""


class CallbackModule(CallbackBase):
    """Dump thread stacks and exit when an execution deadline is reached."""

    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = 'aggregate'
    CALLBACK_NAME = 'timeout'
    CALLBACK_NEEDS_ENABLED = True

    def __init__(self):
        super().__init__()

        deadline_str = os.environ.get('ANSIBLE_TEST_TIMEOUT_DEADLINE')
        dump_dir = os.environ.get('ANSIBLE_TEST_TIMEOUT_DIR')

        if not deadline_str or not dump_dir:
            return

        deadline = datetime.datetime.fromtimestamp(float(deadline_str), tz=datetime.timezone.utc)
        now = datetime.datetime.now(datetime.timezone.utc)
        remaining = deadline - now

        dump_dir_path = pathlib.Path(dump_dir)
        dump_dir_path.mkdir(parents=True, exist_ok=True)

        dump_path = dump_dir_path / f'timeout-{now.isoformat()}-{os.getpid()}.txt'

        self._dump_file = dump_path.open('w')
        self._dump_file.write(f'Command: {shlex.join(sys.argv).replace("\n", " ")}\n')
        self._dump_file.flush()

        if remaining > datetime.timedelta():
            self._dump_file.write(f'Timeout: {remaining} remaining\n')
            self._dump_file.flush()
            faulthandler.dump_traceback_later(remaining.total_seconds(), file=self._dump_file, exit=True)
        else:
            self._dump_file.write(f'Timeout: deadline exceeded by {-remaining}\n')
            self._dump_file.flush()
            faulthandler.dump_traceback(file=self._dump_file)
            sys.exit(1)



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import ansible.constants as C
from ansible.errors import AnsibleParserError
from ansible.module_utils.common.sentinel import Sentinel
from ansible.playbook.block import Block
from ansible.playbook.task import Task
from ansible.utils.display import Display

__all__ = ['TaskInclude']

display = Display()


class TaskInclude(Task):

    """
    A task include is derived from a regular task to handle the special
    circumstances related to the `- include_*: ...` task.
    """

    BASE = frozenset(('file', '_raw_params'))  # directly assigned
    OTHER_ARGS = frozenset(('apply',))  # assigned to matching property
    VALID_ARGS = BASE.union(OTHER_ARGS)  # all valid args
    VALID_INCLUDE_KEYWORDS = frozenset(('action', 'args', 'collections', 'debugger', 'ignore_errors', 'loop', 'loop_control',
                                        'loop_with', 'name', 'no_log', 'register', 'run_once', 'tags', 'timeout', 'vars',
                                        'when'))

    def __init__(self, block=None, role=None, task_include=None):
        super(TaskInclude, self).__init__(block=block, role=role, task_include=task_include)
        self.statically_loaded = False

    @staticmethod
    def load(data, block=None, role=None, task_include=None, variable_manager=None, loader=None):
        ti = TaskInclude(block=block, role=role, task_include=task_include)
        task = ti.check_options(
            ti.load_data(data, variable_manager=variable_manager, loader=loader),
            data
        )

        return task

    def check_options(self, task, data):
        """
        Method for options validation to use in 'load_data' for TaskInclude and HandlerTaskInclude
        since they share the same validations. It is not named 'validate_options' on purpose
        to prevent confusion with '_validate_*" methods. Note that the task passed might be changed
        as a side-effect of this method.
        """
        my_arg_names = frozenset(task.args.keys())

        # validate bad args, otherwise we silently ignore
        bad_opts = my_arg_names.difference(self.VALID_ARGS)
        if bad_opts and task.action in C._ACTION_ALL_PROPER_INCLUDE_IMPORT_TASKS:
            raise AnsibleParserError('Invalid options for %s: %s' % (task.action, ','.join(list(bad_opts))), obj=data)

        if not task.args.get('_raw_params'):
            task.args['_raw_params'] = task.args.pop('file', None)
            if not task.args['_raw_params']:
                raise AnsibleParserError('No file specified for %s' % task.action, obj=data)

        apply_attrs = task.args.get('apply', {})
        if apply_attrs and task.action not in C._ACTION_INCLUDE_TASKS:
            raise AnsibleParserError('Invalid options for %s: apply' % task.action, obj=data)
        elif not isinstance(apply_attrs, dict):
            raise AnsibleParserError('Expected a dict for apply but got %s instead' % type(apply_attrs), obj=data)

        return task

    def preprocess_data(self, ds):
        ds = super(TaskInclude, self).preprocess_data(ds)

        diff = set(ds.keys()).difference(self.VALID_INCLUDE_KEYWORDS)
        for k in diff:
            # This check doesn't handle ``include`` as we have no idea at this point if it is static or not
            if ds[k] is not Sentinel and ds['action'] in C._ACTION_ALL_INCLUDE_ROLE_TASKS:
                if C.INVALID_TASK_ATTRIBUTE_FAILED:
                    raise AnsibleParserError("'%s' is not a valid attribute for a %s" % (k, self.__class__.__name__), obj=ds)
                else:
                    display.warning("Ignoring invalid attribute: %s" % k)

        return ds

    def copy(self, exclude_parent=False, exclude_tasks=False):
        new_me = super(TaskInclude, self).copy(exclude_parent=exclude_parent, exclude_tasks=exclude_tasks)
        new_me.statically_loaded = self.statically_loaded
        return new_me

    def build_parent_block(self):
        """
        This method is used to create the parent block for the included tasks
        when ``apply`` is specified
        """
        apply_attrs = self.args.pop('apply', {})
        if apply_attrs:
            apply_attrs['block'] = []
            p_block = Block.load(
                apply_attrs,
                play=self._parent._play,
                task_include=self,
                role=self._role,
                variable_manager=self._variable_manager,
                loader=self._loader,
            )
        else:
            p_block = self

        return p_block



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from ansible._internal._templating._engine import TemplateEngine
from ansible.errors import AnsibleError
from ansible.playbook.attribute import NonInheritableFieldAttribute
from ansible.playbook.base import FieldAttributeBase
from ansible.utils.vars import validate_variable_name


class LoopControl(FieldAttributeBase):

    loop_var = NonInheritableFieldAttribute(isa='string', default='item', always_post_validate=True)
    index_var = NonInheritableFieldAttribute(isa='string', always_post_validate=True)
    label = NonInheritableFieldAttribute(isa='string')
    pause = NonInheritableFieldAttribute(isa='float', default=0, always_post_validate=True)
    extended = NonInheritableFieldAttribute(isa='bool', always_post_validate=True)
    extended_allitems = NonInheritableFieldAttribute(isa='bool', default=True, always_post_validate=True)
    break_when = NonInheritableFieldAttribute(isa='list', default=list, always_post_validate=True)

    def __init__(self):
        super(LoopControl, self).__init__()

    @staticmethod
    def load(data, variable_manager=None, loader=None):
        t = LoopControl()
        return t.load_data(data, variable_manager=variable_manager, loader=loader)

    def post_validate(self, templar: TemplateEngine) -> None:
        super().post_validate(templar)

        try:
            validate_variable_name(self.loop_var)
        except AnsibleError as ex:
            raise AnsibleError("Invalid 'loop_var'.") from ex

        if self.index_var is not None:
            try:
                validate_variable_name(self.index_var)
            except AnsibleError as ex:
                raise AnsibleError("Invalid 'index_var'.") from ex

    def _post_validate_break_when(self, attr, value, templar):
        """
        break_when is evaluated after the execution of the loop is complete,
        and should not be templated during the regular post_validate step.
        """
        return value



from __future__ import annotations

import contextlib as _contextlib
import dataclasses as _dataclasses
from importlib import resources as _importlib_resources, util as _importlib_util
import inspect as _inspect
import pathlib as _pathlib
import typing as _t

if _t.TYPE_CHECKING:
    from ansible.module_utils.compat.typing import LiteralString


class EmbedManager:
    """Utility class for embedding arbitrary content in an AnsiballZ payload."""
    @classmethod
    def embed(cls, package: LiteralString, resource: LiteralString, /) -> EmbeddedResource:
        """
        Request embedding of the specified `resource` from `package` in the generated AnsiballZ payload.
        The `package` argument must resolve to a Python package beneath `ansible` or `ansible_collections`.
        Relative import-style strings are supported (e.g. `..module_utils.something`).
        The `resource` argument must match the target filename.
        The method returns an `EmbeddedResource` object that must be stored and used to access the embedded content at runtime.

        To enable embedding analysis during payload build, the `embed` module must be imported in one of the following ways:
        `from ansible.module_utils.embed import EmbedManager` (optional alias supported)
        `from ansible.module_utils import embed` (optional alias supported)

        To be properly detected during payload build analysis, the `embed` call must:
        * Occur at the topmost level of a module or module_util.
        * Assign its result to a valid variable.
        * Use only inline literal string values as positional arguments.
        * Use only the as-imported name of the `embed` module or `EmbedManager` type (including import aliases).
        """
        if package.startswith('.'):
            st = _inspect.stack()
            sp = _importlib_util.find_spec(st[1].frame.f_globals['__name__'])
            package = _importlib_util.resolve_name(package, sp.parent)
            # FUTURE: register this value for runtime discovery of "what's embedded here"
        return EmbeddedResource(package, resource)


@_dataclasses.dataclass(frozen=True)
class EmbeddedResource:
    """Wrapper object returned by `EmbedManager.embed()` for runtime access to content embedded in an AnsiballZ payload."""
    package: str
    resource: str

    @property
    def path_context_manager(self) -> _contextlib.AbstractContextManager[_pathlib.Path]:
        """
        Returns a context manager that, once entered, provides a `pathlib.Path` to the embedded content.

        Path validity is only guaranteed until the context manager exits, as the temporarily extracted content is deleted when running from a zip archive.
        """
        return _importlib_resources.path(self.package, self.resource)

    @property
    def python_module_ref(self) -> str:
        """Returns a fully-qualified Python module reference to the embedded content (with the `.py` extension suppressed)."""
        return self.package + "." + str(_pathlib.Path(self.resource).with_suffix(''))
